# Mapping processor following the parser definition
import pandas as pd
import json, sqlite3
import importlib.util
import os
import migration_machine.config as configurations
from pathlib import Path
from datapackage import Package, exceptions
from migration_machine.utils.utils import LoggingUtils
from functools import partial


class Mapper:
    def __init__(self, configs):
        self.configs = configs
        self.aux_modules = {}
        self.logging = LoggingUtils(configs)

    def related_resource(self, fields_source, fields_map):
        """
        Find the relation configuration of the related entity with the source entity
        :param fields_source:
        :param fields_map:
        :return:
        """
        reference = next((ref for ref in fields_source if ref["name"] == fields_map["relation"]), None)
        fields = fields_map["fields"]
        if reference is None:
            return fields, reference
        if reference['reference']['fields'] not in fields_map["fields"]:
            fields.append(reference['reference']['fields'])
        return fields, reference

    def filter_same_name_col(self, source_pk, source_col, other_col):
        """
        Exclude same name columns inside the entities being merged
        :param source_pk:
        :param source_col:
        :param other_col:
        :return:
        """
        other_col_to_consider = list(other_col)

        for sp in source_pk:
            if sp in other_col_to_consider:
                other_col_to_consider.remove(sp)  # in case the PK of the source has the same name of one of the
                # columns of the related entity
        col_to_consider = source_col.difference(other_col_to_consider)
        return col_to_consider

    def merge_relations(self, dir_data, source_datapkg, source, source_entity_name, source_entity_spec, map_relations,
                        format_source):
        """
        Merge the source and the related dataframes applying criteria, aggregations and using the subset of columns
        :param dir_data:
        :param source_datapkg:
        :param source:
        :param source_entity_name:
        :param source_entity_spec:
        :param map_relations:
        :return:
        """
        self.logging.log_generator("starting merge_relations: " + source_entity_name, False)
        source_relations = source_entity_spec['relations'] if 'relations' in source_entity_spec else []
        indicator_types = ["left_only", "right_only", "both"]
        # Expand source table with relations
        for fk in map_relations:
            self.logging.log_generator("merging entities of {}: " .format(fk["relation"]), False)
            filters = None
            multifilters = None
            aggregator = None
            # find relation between two tables, if any
            related_res = self.related_resource(source_relations, fk)
            # related table fields
            fields = related_res[0]

            self.logging.log_generator(fields, False)
            related_entity_spec = source_datapkg.get_resource(fk['relation']).descriptor['schema']
            primary_key = related_entity_spec['primaryKey']
            if related_res[1] is None:
                if isinstance(primary_key, str):
                    fields.append(primary_key)
                else:
                    fields = fields + primary_key

            if "condition" in fk:
                filters = fk["condition"]
            if "multicondition" in fk:
                multifilters = fk["multicondition"]
            if "aggregations" in fk:
                aggregator = fk["aggregations"]

            other = self.pandas_load_dataframe(dir_data, fk['relation'], columns=fields, selection_criteria=filters,
                                          aggregator=aggregator, primary_key=primary_key, format_source=format_source)
            other = other.reset_index(drop=True)
            source = source.reset_index(drop=True)

            # use merge fields from relation or defined explicitly with source_fields/target_fields
            left_on = fk['source_fields'] if 'source_fields' in fk else related_res[1]['fields']
            right_on = fk['target_fields'] if 'target_fields' in fk else related_res[1]['reference']['fields']
            if isinstance(right_on, str):
                right_on = [right_on]
            if 'alias' in fk:
                newCols = []
                for elem in other.columns:
                    newCols.append(fk['alias'] + '_' + elem)
                for index, elem in enumerate(right_on):
                    right_on[index] = fk['alias'] + '_' + elem
                for index, elem in enumerate(fields):
                    fields[index] = fk['alias'] + '_' + elem
                other.columns = newCols

            col_to_consider = self.filter_same_name_col(left_on, source.columns, other.columns)

            if "join" in fk and fk["join"] in indicator_types:
                source = source[col_to_consider].merge(other, left_on=left_on, right_on=right_on, indicator=True)
                source = source[source["_merge"] == fk["join"]]  # d1.index.difference(d2.index)
                del source["_merge"]
            else:
                source = source[col_to_consider].merge(other, how='left', left_on=left_on, right_on=right_on)

            if multifilters is not None and len(multifilters) > 0:
                source = source.query(" and ".join(multifilters))
            self.logging.log_generator("merged dataframe shape: " + str(source.shape), False)
        # source.to_csv('tmp.csv', index=False, encoding="ISO-8859-1")
        return source

    def load_parquet(self, dir_data, source_entity_name, engine="pyarrow", columns=None):
        if columns: columns = list(dict.fromkeys(columns))
        source_entity = pd.read_parquet("{}/{}_{}.parquet".format(dir_data, source_entity_name, source_entity_name),
                                        engine=engine, columns=columns)
        return source_entity

    def load_csv(self, dir_data, source_entity_name, encoding="ISO-8859-1", sep=",", dtype=str, usecols=None):
        source_entity = pd.read_csv("{}/{}_{}.csv".format(dir_data, source_entity_name, source_entity_name),
                                    encoding=encoding, sep=sep, dtype=dtype, usecols=usecols)
        return source_entity

    def load_sqlite(self, dir_data, source_entity_name, selection_criteria=None, columns=None):
        conn = sqlite3.connect("{}/{}_{}.sqlite".format(dir_data, source_entity_name, source_entity_name))
        filtering = " 1 "
        if selection_criteria and selection_criteria is not None:
            filtering = selection_criteria
            type_filter = type(filtering)
            if type_filter is list:
                filtering = " and ".join(filtering)
            filtering = filtering.replace("[", "(").replace("]", ")")
        q = "SELECT {} FROM {} WHERE {}".format(columns, "{}_{}".format(source_entity_name, source_entity_name),
                                                filtering)
        return pd.read_sql_query(q, conn)

    def pandas_load_dataframe(self, dir_data, source_entity_name, selection_criteria=None, columns=None, aggregator=None,
                              primary_key=None, format_source="parquet"):
        """
        Load the source files into a pandas dataframe applying the specified criteria, columns and aggregation
        :param dir_data:
        :param source_entity_name:
        :param selection_criteria:
        :param columns:
        :param aggregator:
        :param primary_key:
        :return:
        """
        if format_source == "csv":
            loader = partial(self.load_csv, dir_data, source_entity_name, encoding="ISO-8859-1", sep=",", dtype=str)
            if columns is not None:
                source_entity = loader(usecols=columns)
            else:
                source_entity = loader()
        elif format_source == "sqlite":
            loader = partial(self.load_sqlite, dir_data, source_entity_name, selection_criteria=selection_criteria)
            if columns is not None:
                source_entity = loader(columns=", ".join(columns))
            else:
                source_entity = loader(columns="*")
        else:
            loader = partial(self.load_parquet, dir_data, source_entity_name, engine="pyarrow")
            if columns is not None:
                source_entity = loader(columns=columns)
            else:
                source_entity = loader()
        source_entity.reset_index(inplace=True, drop=True)
        source_entity.drop_duplicates(inplace=True)

        # filter dataframe rows based on criteria
        if selection_criteria and selection_criteria is not None and format_source != "sqlite":
            filtering = selection_criteria
            type_filter = type(filtering)
            if type_filter is list:
                filtering = " and ".join(filtering)
            self.logging.log_generator("filtering dataframe : " + source_entity_name, False)
            self.logging.log_generator(filtering, False)
            source_entity = source_entity.query(filtering)
        # apply aggregations if necessary
        if aggregator is not None and primary_key is not None:
            custom_function = self.generate_custom_func_relation(aggregator)
            reduced_pk = primary_key if columns is None else list(set(primary_key) & set(columns))
            # reduced_pk.remove(aggregator)
            reduced_pk = list(set(reduced_pk).difference(set(aggregator.keys())))
            source_entity = source_entity.groupby(reduced_pk, as_index=False).aggregate(custom_function)

        return source_entity

    def field_mapping(self, mapper, related_filtered, source_entity_spec):
        """
        Apply functions, aggregations and mapping of the source fields into the target ones
        :param mapper:
        :param related_filtered:
        :param source_entity_spec:
        :return:
        """
        self.logging.log_generator("Starting final field mapping...", False)
        for fld_map in mapper['fieldMappings']:
            self.logging.log_generator(fld_map, False)
            # related_filtered.fillna("", inplace=True)
            temp = related_filtered
            if "aggregations" in fld_map:
                related_filtered[fld_map["field"]] = temp.groupby(source_entity_spec['primaryKey'], as_index=False).agg(
                    fld_map["aggregations"])
            elif "function" in fld_map:
                custom_function = self.generate_custom_func(fld_map["function"])
                self.logging.log_generator(custom_function, False)
                if len(temp) > 0:
                    related_filtered[fld_map["field"]] = temp.apply(lambda x: eval(custom_function, self.aux_modules, {"x": x}), axis=1)
                elif fld_map["field"] not in temp.columns:
                    related_filtered[fld_map["field"]] = ""
            elif "source" in fld_map:
                f = fld_map["source"] if "alias" not in fld_map else (fld_map["alias"] + '_' + fld_map["source"])
                related_filtered[fld_map["field"]] = temp[f]
            else:
                related_filtered[fld_map["field"]] = eval(fld_map["value"])

    def generate_custom_func(self, custom_function):
        """
        Transform the parameters of the custom function into the format applicable for pandas
        :param custom_function:
        :return:
        """
        param_start = custom_function.find("(")
        param_end = custom_function.find(")")
        params = custom_function[param_start + 1:param_end]
        if params.strip() == '':
            return custom_function[0:param_start] + '()'
        else:
            #        par = ["x." + utils.normalize_field_name(fld.strip()) for fld in params.split(",")]
            par = ["x." + fld.strip() if "'" not in fld else fld for fld in params.split(",")]
            func = custom_function[0:param_start] + "({})".format(",".join(par))
            return func

    def generate_custom_func_relation(self, aggregations):
        """
        Generates the dictionary of the custom aggregation functions to be applied on columns
        :param aggregations:
        :return:
        """
        standard_func = ["mean", "sum", "count", "size", "std", "var", "sem", "describe", "first", "last", "nth", "min",
                         "max"]
        custom_functions = {
            col: eval("aux." + func, self.aux_modules) if func not in standard_func else func
            for col, func in aggregations.items()
            }
        return custom_functions

    def execute_mapping_file(self, mapper_specifications_file, index=None):
        """
        Execute main mapping routine
        :param index:
        :param mapper_specifications_file:
        :return:
        """
        self.logging.log_generator("Starting processor", False)
        with open(mapper_specifications_file) as file:
            spec = json.load(file)
        self.import_modules(spec)
        # read source and target data packages
        source_datapkg = Package(self.configs.domain["domain_settings_path"] + spec['sourcePackage'] + '.json')
        target_datapkg = Package(self.configs.domain["domain_settings_path"] + spec['targetPackage'] + '.json')
        input_dir = self.configs.mapper["input_dir"]
        output_dir = self.configs.mapper["output_dir"]
        format_source = self.configs.mapper["format_source"]

        # process mappings
        if index is not None:
            mapping = next((mapping for mapping in spec['mappings'] if mapping["mapping_index"] == int(index)), None)
            self.run_single_mapping(mapping, input_dir, output_dir, format_source, source_datapkg, target_datapkg)
        else:
            for mapping in spec['mappings']:
                self.run_single_mapping(mapping, input_dir, output_dir, format_source, source_datapkg, target_datapkg)
        self.logging.log_generator("Ending processor", True)

    def run_single_mapping(self, mapping, dir_data, dir_output, format_source, source_datapkg, target_datapkg):
        """
        Execute single mapping
        :param dir_data:
        :param dir_output:
        :param format_source:
        :param mapping:
        :param source_datapkg:
        :param target_datapkg:
        :return:
        """
        if mapping is None:
            self.logging.log_generator_error("Mapping index does not exist ")
        source_entity_name = mapping['source']
        target_entity_name = mapping['target']
        mapping_index = mapping["mapping_index"]
        relations = mapping['relations']
        selection_criteria = mapping['selectionCriteria']
        self.logging.log_generator(
            "inside mapping: " + str(source_entity_name) + " -> " + str(target_entity_name) + " - " + str(
                mapping_index),
            False)
        source_entity_spec = source_datapkg.get_resource(source_entity_name).descriptor['schema']
        source_entity_df = self.pandas_load_dataframe(dir_data, source_entity_name, selection_criteria=selection_criteria,
                                                 format_source=format_source)
        # TODO use Dask to create chunks of parallel dataframes
        related_filtered = self.merge_relations(dir_data, source_datapkg, source_entity_df, source_entity_name,
                                           source_entity_spec, relations, format_source)
        related_filtered = related_filtered.apply(lambda x: x if x is not None else "")
        # map merged dataframe to the target columns
        self.field_mapping(mapping, related_filtered, source_entity_spec)
        present_target_flds = [fld["field"] for fld in mapping['fieldMappings']]
        related_filtered = related_filtered[present_target_flds]
        related_filtered = related_filtered.drop_duplicates()
        # skip validation of target dataframe
        dspec = target_datapkg.get_resource(target_entity_name).descriptor
        dirname = os.path.dirname(dir_output + '/' + dspec['path'])
        Path(dirname).mkdir(parents=True, exist_ok=True)
        filename = "{}/{}_{}.{}".format(dir_output, dspec['path'], mapping_index, dspec['format'])
        related_filtered.to_csv(filename, index=False, encoding="ISO-8859-1")

    def import_modules(self, mapper_specifications):
        """
        Dynamically import auxiliary modules
        :param mapper_specifications:
        :return:
        """
        if 'imports' in mapper_specifications:
            for module_name, module_path in mapper_specifications['imports'].items():
                self.logging.log_generator("Dynamically importing module: {}".format(module_name), False)
                path = self.configs.domain["domain_aux_path"]
                _modspec = importlib.util.spec_from_file_location(module_name, os.path.join(path, module_path))
                module = importlib.util.module_from_spec(_modspec)
                _modspec.loader.exec_module(module)

                tables, domains = module.populate_tables(path)
                module.tables = tables
                module.domains = domains
                self.aux_modules[module_name] = module

    def run_mapping_processor(self):
        """
        Run all mappings for each dataset present in the settings folder of the domain
        :return:
        """
        datasets_folders = os.listdir(self.configs.domain["domain_settings_path"])
        for folder in datasets_folders:
            if os.path.isdir(os.path.join(self.configs.domain["domain_settings_path"], folder)):
                mapper_specifications_file = os.path.join(self.configs.domain["domain_settings_path"], folder,
                                                          "migration_" + folder + '_result.json')
                self.execute_mapping_file(mapper_specifications_file)


if __name__ == "__main__":
    config = configurations.Config("examples/SIB2SICRA/transform.ini")
    mapper = Mapper(config)
    #mapper.run_mapping_processor()
    mapper.execute_mapping_file(os.path.join(config.domain["domain_settings_path"], "ANAGDI", "migration_ANAGDI_result.json"))
