# Mapping processor following the parser definition
import json
import os
import pandas as pd
import numpy as np
import migration_machine.config as configurations
from migration_machine.utils.utils import LoggingUtils


class Merger:
    def __init__(self, configs):
        self.configs = configs
        self.logging = LoggingUtils(configs)

    def merge_table(self, path, pk, fields, dir, output):
        """
        Merge all partially produced subsets
        :param path:
        :param pk:
        :param fields:
        :param dir:
        :param output:
        :return:
        """
        merged = None
        try:
            self.logging.log_generator("Starting merger of dataset {}...".format(path), False)
            if not isinstance(pk, list): pk = [pk]
            file_pre = path[path.rfind('/') + 1:] if path.rfind('/') >= 0 else path
            only_files = [f for f in os.listdir(dir) if os.path.isfile(os.path.join(dir, f)) and f.find(file_pre) == 0]
            if len(only_files) == 0: return

            os.makedirs(os.path.dirname(os.path.abspath(output + '/' + path)), exist_ok=True)

            if len(only_files) == 1:
                partial_dataset = pd.read_csv(dir + '/' + only_files[0], dtype=str, keep_default_na=False)
                partial_dataset = self.rename_headers(partial_dataset)
                partial_dataset.to_csv(output + '/' + path)
                return

            for file in only_files:
                partial_dataset = pd.read_csv(dir + '/' + file, dtype=str, keep_default_na=False, encoding='latin1')
                if len(partial_dataset.index) == 0: continue
                partial_dataset = self.rename_headers(partial_dataset)

                if merged is None:
                    merged = partial_dataset
                    self.logging.log_generator("Merging {entity}: with columns {columns} ".format(entity=file, columns=merged.columns),
                                               False)
                else:
                    shared_cols = merged.columns.intersection(partial_dataset.columns)
                    shared_cols = list(set(shared_cols) - set(pk))
                    self.logging.log_generator(
                        "Merging {entity}: with columns {columns} ".format(entity=file, columns=partial_dataset.columns), False)
                    merged = merged.merge(partial_dataset, on=pk, how='outer')
                    #TODO calc exec time of merge with / without conflict validation
                    for shared_col in shared_cols:
                        print(file)
                        print(shared_col)
                        merged['indicate_conflict'] = np.where(
                            (merged["{}_x".format(shared_col)] != merged["{}_y".format(shared_col)])
                                & (~merged["{}_x".format(shared_col)].isna())
                                & (~merged["{}_y".format(shared_col)].isna()),
                            "conflict",
                            merged["{}_x".format(shared_col)])
                        if len(merged[merged['indicate_conflict'] == "conflict"]) > 0:
                            self.logging.log_generator_error(
                                'CONFLICTS found for {path} in the file: {file}'.format(path=path, file=file), False)
                            #raise Exception(
                            #    'CONFLICTS found for {path} in the file: {file}'.format(path=path, file=file))
                        merged.rename(columns={"{}_x".format(shared_col): shared_col})
                        merged.drop(["indicate_conflict", "{}_y".format(shared_col)], 1, inplace=True)
            merged = merged.drop_duplicates()
            merged = self.normalize_table(merged, fields)
            merged = merged.sort_values(by=pk)
        finally:
            merged.to_csv(output + '/' + path, index=False)
            self.logging.log_generator("Ending merger of dataset {}...".format(path))

    def rename_headers(self, table):
        names = {}
        for col in table.columns:
            if col[0:2] == 't_':
                names[col] = col[2:]
        table = table.rename(columns=names)
        return table

    def normalize_table(self, table, fields):
        header = table.columns
        res = pd.DataFrame()
        for field in fields:
            if field["name"] in header:
                res[field["name"]] = table[field["name"]]
            else:
                res[field["name"]] = ''
        return res

    def run_merger_all(self):
        datasets_folders = os.listdir(self.configs.domain["domain_settings_path"])
        for dataset in datasets_folders:
            self.run_merger_single_dataset(dataset)

    def run_merger_single_dataset(self, dataset):
        """
        Merge the partially produced files of the single dataset
        :param dataset:
        :return:
        """
        spec_file = self.configs.domain["domain_settings_path"] + self.configs.domain['package_target'] + '.json'
        output_dir = self.configs.merger["output_dir"]
        partial_datasets_dir = self.configs.merger['partial_datasets_dir']
        with open(spec_file) as file:
            specifications_target = json.load(file)
        if os.path.isdir(os.path.join(self.configs.domain["domain_settings_path"], dataset)):
            resource = next(
                (resource for resource in specifications_target['resources'] if resource["name"] == dataset),
                None)
            if resource is not None:
                self.merge_table(resource['path'], resource['schema']['primaryKey'], resource['schema']['fields'],
                                 partial_datasets_dir, output_dir)


if __name__ == "__main__":
    config = configurations.Config("examples/SIB2SICRA/transform.ini")
    merger = Merger(config)
    merger.run_merger_single_dataset("ANAGDG")
    # merger.run_merger_all()
