import sys
import argparse
import json
from datapackage import Package, exceptions
import pandas as pd
from processor.migration_machine.common import logging_util
from pyarrow import csv, parquet
import dask.dataframe as dask


class Converter(object):
    """
    Necessary utilities for validating and loading into pandas the csv files and then converting into parquet files
    """
    def __init__(self, spec, dir):
        self.dataframes = None
        self.spec = spec
        self.dir = dir
        self.source_datapkg = Package(self.spec, base_path=self.dir)

    def __repr__(self):
        return 'Converter({0.spec!r}, {0.dir!r})'.format(self)

    def __str__(self):
        return "({0.spec!s}, {0.dir!s})".format(self)

    def import_csv_files_into_pandas(self):
        """
        Read source data packages based on package definition, including also data validation.
        """
        try:
            logging_util.log_generator("before loading source datapackage into pandas", True)
            self.dataframes = self.source_datapkg.save(storage='pandas')
            logging_util.log_generator("after loading source datapackage into pandas", True)
        except exceptions.TableSchemaException as exception:
            print(exception)
            for error in exception.errors:
                print(error)
            sys.exit()

    def convert_csv_2parquet_with_pandas(self):
        """
        Convert pandas dataframes into parquet
        :return:
        """
        logging_util.log_generator("before converting into pandas", True)
        for resource in self.source_datapkg.resource_names:
            path = self.dir + "/" + self.source_datapkg.get_resource(resource).descriptor['path']
            name = self.source_datapkg.get_resource(resource).descriptor['name']
            print(path)
            table = pd.read_csv(path, encoding='ISO-8859-1', low_memory=False)
            table.to_parquet('{}_pandas.parquet'.format(self.dir + "/SIB/" + name))
        logging_util.log_generator("after converting into pandas", True)

    def convert_csv_2parquet_with_pyarrow(self):
        """
        Convert csv files to parquet using pyarrow
        :param parquet_file:
        :return:
        """
        logging_util.log_generator("before converting using pyarrow", True)
        for resource in self.source_datapkg.resource_names:
            path = self.dir + "/" + self.source_datapkg.get_resource(resource).descriptor['path']
            name = self.source_datapkg.get_resource(resource).descriptor['name']
            print(path)
            table = csv.read_csv(path)
            parquet.write_table(table, '{}_pyarrow.parquet'.format(self.dir + "/SIB/" + name))
        logging_util.log_generator("after converting using pyarrow", True)

    def convert_csv_2parquet_with_dask(self):
        """
        Convert csv files to parquet using dask
        :param parquet_file:
        :return:
        """
        logging_util.log_generator("before converting using dask", True)
        for resource in self.source_datapkg.resource_names:
            path = self.dir + "/" + self.source_datapkg.get_resource(resource).descriptor['path']
            name = self.source_datapkg.get_resource(resource).descriptor['name']
            print(path)
            table = dask.read_csv(path, encoding='ISO-8859-1', assume_missing=True, dtype={
                'status_record': 'object',
                'tipo_codice_alternativo': 'object',
                'tipo__legame_anagr': 'object',
                'forz_soglia_amm_socio': 'object',
                'cag_sottoscrittore': 'object',
                'codice_rapporto': 'object',
                'codice_riferimento_o_perazione': 'object',
                'filler_2': 'object',
                'motivazione_raccolta_quest': 'object',
                'presenza_del_titolar_e_effettivo': 'object',
                'stato_questionario': 'object'
            }, low_memory = False)
            table.to_parquet('{}_dask.parquet'.format(self.dir + "/SIB/" + name))
        logging_util.log_generator("after converting using dask", True)

parser = argparse.ArgumentParser(description='Map datapackage')
parser.add_argument('-s', '--spec', help='Mapping specification file')
parser.add_argument('-d', '--dir', default='.', help='working data directory')
args = vars(parser.parse_args())
spec_file = args['spec']
directory = args['dir']

with open(spec_file) as f: spec = json.load(f)
logging_util.log_generator("Starting parser")
c = Converter(spec, directory)
c.convert_csv_2parquet_with_pandas()
c.convert_csv_2parquet_with_dask()
c.convert_csv_2parquet_with_pyarrow()

logging_util.log_generator("Ending parser")