# PARSING FOR SIB ARCHIVES
# Assuming 
# - presence of TRACCIA_TRACCIA file in the specified directory
# - 7 columns with folder, table short name, table description, field number, field name, datatype, format
# generates data package JSON definition saving it in a specified output file
import sqlite3
import os, csv, json
import pandas as pd
import pyarrow.parquet as pq
import pyarrow as pa
from migration_machine.utils.utils import LoggingUtils, normalize_field_name
import migration_machine.config as configurations


class SourceParser:
    def __init__(self, configs):
        self.config = configs
        self.logging = LoggingUtils(configs)

    def extract_type_def(self, format, mask, field):
        """
        Extract field type
        :param format:
        :param mask:
        :param field:
        """
        if format == 'D':
            field['type'] = 'date'
            field['format'] = '%d.%m.%Y'
        elif format == 'T':
            field['type'] = 'time'
            field['format'] = 'default'
        elif format == 'A':
            field['type'] = 'string'
            field['constraints'] = {'maxLength': int(int(mask) / 100000), 'unique': False}
        elif format == 'N':
            floats = int(mask) % 100000
            if floats == 0:
                field['type'] = 'integer'
            else:
                field['type'] = 'number'
                field['decimalChar'] = ','

    def load_df_entire(self, data_file_path, names, format_convert, dest_data_file_path, data_file_name):
        """
        Load into pandas dataframe the entire data file
        :param data_file_name:
        :param dest_data_file_path:
        :param data_file_path:
        :param names:
        :param format_convert:
        """
        data_file = pd.read_csv(data_file_path, sep='|', header=None, encoding="ISO-8859-1", skipinitialspace=True,
                                dtype=str, quoting=csv.QUOTE_NONE,
                                keep_default_na=False)
        data_file = data_file.drop([0, len(names) + 1], axis=1)
        data_file.columns = names
        data_file.fillna("", inplace=True)
        data_file = data_file.apply(lambda x: x.str.strip() if x.dtype == "object" else x)
        for col in names:
            data_file[col] = data_file[col].astype(str)
        if format_convert == "parquet":
            data_file.columns = data_file.columns.astype(str)
            data_file.to_parquet(dest_data_file_path + '.parquet', index=False)
        elif format_convert == "sqlite":
            db = sqlite3.connect(dest_data_file_path + ".sqlite")
            data_file.to_sql(data_file_name, db, if_exists="append")
            db.close()
        else:
            data_file.to_csv(dest_data_file_path + '.csv', index=False, encoding="ISO-8859-1")

    def load_df_chunked(self, data_file_path, names, name_key, format_convert, dest_data_file_path, data_file_name):
        """
        Read and write chunks of data into pandas dataframes
        :param data_file_path:
        :param names:
        :param name_key:
        :param format_convert:
        """
        chunk_size = 500000
        if len(names) < 60:
            chunk_size = 1000000
        chunks = pd.read_csv(data_file_path, sep='|', header=None, encoding="ISO-8859-1", skipinitialspace=True,
                             dtype=str, quoting=csv.QUOTE_NONE,
                             chunksize=chunk_size, iterator=True, keep_default_na=False)
        pqwriter = None
        db = None
        for i, chunk in enumerate(chunks):
            chunk = chunk.drop([0, len(names) + 1], axis=1)
            chunk.columns = names
            chunk.fillna("", inplace=True)
            chunk.columns = chunk.columns.astype(str)
            chunk = chunk.apply(lambda x: x.str.strip() if x.dtype == "object" else x)
            if format_convert == "parquet":
                table = pa.Table.from_pandas(chunk)
                if i == 0:
                    # create a parquet write object giving it an output file
                    pqwriter = pq.ParquetWriter(dest_data_file_path + '.parquet', table.schema)
                pqwriter.write_table(table)
            elif format_convert == "sqlite":
                if i == 0:
                    db = sqlite3.connect(dest_data_file_path + ".sqlite")
                chunk.to_sql(data_file_name, db, if_exists="append")
            else:
                chunk.to_csv(dest_data_file_path + '.csv', mode='a', index=False, header=False, encoding="ISO-8859-1")
        # close the parquet writer
        if pqwriter:
            pqwriter.close()
        if db:
            db.close()

    def parse_source(self, packagename, output, datapackage, dfg, transform, format_convert, domain_data_path,
                     regenerate_schema):
        self.logging.log_generator("Starting parser source ...", False)
        dest_data_file_path = domain_data_path + "/" + format_convert
        os.makedirs(dest_data_file_path, exist_ok=True)
        for key in dfg.groups:
            data_file_name = key[1] + '_' + key[1]
            names = self.output_schema(data_file_name, datapackage, dfg, format_convert, key, packagename)
            data_file_path = domain_data_path + '/' + data_file_name
            self.output_format(data_file_name, data_file_path, dest_data_file_path, format_convert, key, names,
                               transform)

        if regenerate_schema:
            with open(output, 'w') as outfile:
                json.dump(datapackage, outfile, indent=4)
        self.logging.log_generator("Ending parser source...")

    def output_schema(self, data_file_name, datapackage, dfg, format_convert, key, packagename):
        resource = {
            'path': packagename + '/' + data_file_name + '.{}'.format(format_convert),
            'format': format_convert,
            'name': key[1],
            'description': key[2][0:-2].strip(),
            'encoding': 'ISO-8859-1',
            'schema': {
                'primaryKey': '', 'fields': []
                }
            }
        datapackage['resources'].append(resource)
        group = dfg.get_group(key)
        names = []
        for idx, prop in group.iterrows():
            name = normalize_field_name(prop[4])
            field = {
                'name': name,
                'description': prop[4] + ' (' + str(prop[3]) + ')'
                }
            self.extract_type_def(prop[5], prop[6], field)
            resource['schema']['fields'].append(field)
            if name in names:
                name += str(prop[3])
                field['name'] = name
            names.append(name)
        return names

    def output_format(self, data_file_name, data_file_path, dest_data_file_path, format_convert, key, names, transform):
        if transform and os.path.exists(data_file_path):
            if os.path.getsize(data_file_path) > 0:
                if os.path.getsize(data_file_path) > 1500000000:  # 1,5 GB
                    self.load_df_chunked(data_file_path, names, key[1], format_convert,
                                         dest_data_file_path + '/' + data_file_name, data_file_name)
                else:
                    self.load_df_entire(data_file_path, names, format_convert,
                                        dest_data_file_path + '/' + data_file_name, data_file_name)

    def read_config(self, configs):
        packagename = configs.domain["package_source"]
        source_schema = "/" + configs.parser_source["original_schema_file"]
        regenerate_schema = eval(configs.parser_source["regenerate_schema"])
        domain_data_path = configs.domain["domain_data_path"] + configurations.CONFIG_DOMAIN_SOURCE_FOLDER
        source_schema_path = domain_data_path + source_schema
        output = configs.domain["domain_settings_path"] + packagename + ".json"
        transform = eval(configs.parser_source["transform"])
        format_convert = configs.parser_source['format']
        return domain_data_path, format_convert, output, packagename, regenerate_schema, source_schema_path, transform

    def run_parser(self, configs):
        domain_data_path, format_convert, output, packagename, regenerate_schema, source_schema_path, transform = self.read_config(
            configs)
        df = pd.read_fwf(source_schema_path, header=None, widths=[10, 10, 50, 10, 50, 5, 10], encoding="ISO-8859-1")
        dfg = df.groupby([0, 1, 2])
        datapackage = {'name': packagename, 'resources': []}
        self.parse_source(packagename, output, datapackage, dfg, transform, format_convert, domain_data_path,
                          regenerate_schema)


if __name__ == "__main__":
    config = configurations.Config("examples/SIB2SICRA/transform.ini")
    parser = SourceParser(config)
    parser.run_parser(config)
