# PARSING FOR SICRA ARCHIVES
# Assuming 
# - presence of description FW files named as QDDSFFD.<name>
# - 7 columns with field id, datatype, length, position, field description, field id, field number
# generates data package JSON definition saving it in a specified output file
import pandas as pd
import json
import os
import migration_machine.config as configurations
from migration_machine.utils.utils import normalize_field_name, LoggingUtils


class TargetParser:
    def __init__(self, configs):
        self.configs = configs
        self.logging = LoggingUtils(configs)

    def extract_type_def(self, format, len, field):
        # 'type': 'number' if prop[5] == 'N' else 'date' if prop[5] == 'D' else 'time' if prop[5] == 'T' else 'string'
        if format == 'P':
            field['type'] = 'boolean'
            field['trueValues'] = ['1', 'S']
            field['falseValues'] = ['0', 'N']
        elif format == 'D':
            field['type'] = 'date'
            field['format'] = '%d.%m.%Y'
        elif format == 'T':
            field['type'] = 'time'
            field['format'] = 'default'
        elif format == 'A':
            field['type'] = 'string'
            field['constraints'] = {'maxLength': len}
        elif format == 'N':
            field['type'] = 'number'

    def parse_target(self, packagename, path, output, transform, regenerate_schema):
        self.logging.log_generator("Starting parser target...", False)
        _, _, filenames = next(os.walk(path), (None, None, []))
        datapackage = {'name': packagename, 'resources': []}
        for filename in filenames:
            df = pd.read_fwf(path + '/' + filename, skiprows=1, header=None, widths=[12, 1, 6, 8, 21, 21, 34],
                             encoding="ISO-8859-1")
            name = filename.split('.')[1]
            resource = {
                'path': packagename + '/' + name + '.FILE', 'name': name, 'format': 'csv', 'description': filename,
                'schema': {
                    'primaryKey': '', 'fields': []
                    }
                }
            datapackage['resources'].append(resource)
            names = []
            for idx, prop in df.iterrows():
                field = {
                    'name': normalize_field_name(prop[4]),
                    'description': prop[4] + ' (' + str(prop[0]) + ')'
                    }
                self.extract_type_def(prop[1], prop[2], field)
                resource['schema']['fields'].append(field)
                names.append(field['name'])

            data_file_path = path + '/' + name
            if transform and os.path.exists(data_file_path):
                if os.path.getsize(data_file_path) > 0:
                    data_file = pd.read_csv(data_file_path, sep=';', header=None, encoding="ISO-8859-1",
                                            skipinitialspace=True,
                                            dtype=str)
                    data_file.to_csv(data_file_path + '.csv', header=names, index=False, encoding="ISO-8859-1")
        if regenerate_schema:
            with open(output, 'w') as outfile:
                json.dump(datapackage, outfile, indent=4)
        self.logging.log_generator("Ending parser target...")

    def run_parser_target(self, config):
        packagename = config.domain["package_target"]
        domain_data_path = config.domain["domain_data_path"] + configurations.CONFIG_DOMAIN_TARGET_FOLDER
        output = config.domain["domain_settings_path"] + packagename + ".json"
        transform = eval(config.parser_target["transform"])
        regenerate_schema = eval(config.parser_target["regenerate_schema"])
        self.parse_target(packagename, domain_data_path, output, transform, regenerate_schema)


if __name__ == "__main__":
    config = configurations.Config("examples/SIB2SICRA/transform.ini")
    parser = TargetParser(config)
    parser.run_parser_target(config)
