# PARSING FOR ONTOLOGY-BASED MIGRATION DEFINITIONS
# generates migration JSON definition saving it in a specified output file
import json, os
from migration_machine.utils.utils import LoggingUtils, normalize_field_name
import migration_machine.config as configurations


class Mapping:
    source = None
    target = None
    selectionCriteria = []
    relations = []
    fieldMappings = []

    def __init__(self, source, target, selectionCriteria, relations, fieldMappings):
        self.source = source
        self.target = target
        self.selectionCriteria = selectionCriteria
        self.relations = relations
        self.fieldMappings = fieldMappings


class Relation:
    relation = None
    join = 'both'
    fields = []
    aggregations = []
    condition = []

    def __init__(self, relation, join, fields, aggregations, condition):
        self.relation = relation
        if join is not None:
            self.join = join
        self.fields = fields
        self.aggregations = aggregations
        self.condition = condition


class Converter:
    def __init__(self, configs):
        self.config = configs
        self.logging = LoggingUtils(configs)

    # Extract main source table for mapping: expect to have only one
    def extract_source(self, m, s_spec):
        masters = set()
        if 'source_key' in m:
            for k in m['source_key']:
                masters.add(k['table'])
        else:
            tables = m['source_tables']
            reverse_rel_dict = {}
            rel_dict = {}
            for table in s_spec['resources']:
                if 'relations' in table['schema']:
                    rel_dict[table['name']] = table['schema']['relations']
                    for relation in table['schema']['relations']:
                        reverse_rel_dict[relation['reference']['resource']] = table['name']
            for table in tables:
                if table.startswith('not_'):
                    table = table[4:]
                master = self.find_parent(table, rel_dict, reverse_rel_dict)
                if master is not None:
                    masters.add(master)
                else:
                    masters.add(table)
        if len(masters) > 1:
            self.logging.log_generator_error(
                'INVALID MAPPING: non-unique master source table: {}'.format(','.join(masters)), False)
            raise Exception('INVALID MAPPING: non-unique master source table: {}'.format(','.join(masters)))

        return ''.join(masters)

    def find_parent(self, t, rel_dict, reverse_rel_dict):
        if t not in reverse_rel_dict:
            return t
        return self.find_parent(reverse_rel_dict[t], rel_dict, reverse_rel_dict)

    # Extract main target table for mapping: expect to have only one
    def extract_target(self, tables, t_spec):
        if len(tables) > 1:
            self.logging.log_generator_error('INVALID MAPPING: More than one target table', False)
            raise Exception('INVALID MAPPING: More than one target table')
        return tables[0]

    # extract table name with alias if present
    def get_table_name(self, d):
        res = d['table'] + '_AS_' + d['as'] if 'as' in d else d['table']
        return res

    # extract table name with alias if present
    def get_table_name_rev(self, n):
        res = n.split('_AS_')
        return res

    # prepare local table mapping conditions
    def get_or_create_conditions_for_rel(self, table, rel_conditions, s_spec):
        if table not in rel_conditions:
            rel_conditions[table] = []
            # NOT NEEDED: added automatically during mapping ?
            # main = next(t for t in s_spec['resources'] if x['name'] == table)
            # if main['schema']['relations'] != None:
            #     rel = next(r for r in main['schema']['relations'] if r['reference']['resource'] == table, None)
            #     if rel != None:
            #         exp = ''
            #         rel_conditions[table].append(exp)
        return rel_conditions[table]

    # extract table and field names from the conditions recursively
    def extract_table_rec(self, arg):
        t = type(arg)
        res = []
        if t is list:
            for e in arg:
                res = res + self.extract_table_rec(e)
        elif t is dict and 'table' in arg:
            return [{'table': arg['table'], 'field': arg['field']}]
        elif t is dict:
            if 'arg0' in arg:
                res = res + self.extract_table_rec(arg['arg0'])
            if 'arg1' in arg:
                res = res + self.extract_table_rec(arg['arg1'])
        return res

    # convert value definition to expression
    def to_val(self, val, s_spec):
        t = type(val)
        if t is list:
            res = set()
            for e in val:
                res.add(self.to_val(e, s_spec))
            return '[' + ','.join(res) + ']'
        elif t is dict:
            if 'aggregator' in val:
                self.logging.log_generator_error('NON SUPPORTED MAPPING: aggregator in condition', False)
                raise Exception('NON SUPPORTED MAPPING: aggregator in condition')
            else:
                return normalize_field_name(val['field'])
        elif t is str:
            return "'" + val + "'"
        else:
            return val

    # extract condition expression
    def extract_cond(self, cond, s_spec):
        operator = cond['operator']
        if operator == 'or':
            return '(' + self.extract_cond(cond['arg0'], s_spec) + ') or (' + self.extract_cond(cond['arg1'],
                                                                                                s_spec) + ')'
        if operator == 'and':
            return '(' + self.extract_cond(cond['arg0'], s_spec) + ') and (' + self.extract_cond(cond['arg1'],
                                                                                                 s_spec) + ')'

        arg0 = self.to_val(cond['arg0'], s_spec)
        arg1 = self.to_val(cond['arg1'], s_spec) if 'arg1' in cond else None
        if operator == 'eq':
            return arg0 + ' == ' + arg1
        elif operator == 'neq':
            return arg0 + ' != ' + arg1
        elif operator == 'oneOf':
            return arg0 + ' in ' + arg1
        elif operator == 'notOneOf':
            return arg0 + ' not in ' + arg1
        elif operator == 'range':
            if 'min' in cond and 'max' in cond:
                min_v = self.to_val(cond['min'], s_spec)
                max_v = self.to_val(cond['max'], s_spec)
                return arg0 + ' <= ' + max_v + ' and ' + arg0 + ' >= ' + min_v
            else:
                min_v = self.to_val(cond['arg1'][0], s_spec)
                max_v = self.to_val(cond['arg1'][1], s_spec)
                return arg0 + ' <= ' + max_v + ' and ' + arg0 + ' >= ' + min_v
        elif operator == 'lt':
            return arg0 + ' < ' + arg1
        elif operator == 'lte':
            return arg0 + ' <= ' + arg1
        elif operator == 'gt':
            return arg0 + ' > ' + arg1
        elif operator == 'gte':
            return arg0 + ' >= ' + arg1

    # process a single condition: return a tuple with the tables used and the condition expression to check
    def process_condition(self, pre, s_spec):
        if 'key' in pre:
            return (map(lambda k: {'table': k['table'], 'field': k['field']}, pre['key']), None)
        pre_tables = self.extract_table_rec(pre['arg0'])
        if 'arg1' in pre:
            pre_tables += self.extract_table_rec(pre['arg1'])
        exp = self.extract_cond(pre, s_spec)
        return (pre_tables, exp)

    def extract_mapping_value(self, v_array, res, main_table, version):
        if len(v_array) > 1:
            res['function'] = 'aux.CONCAT(' + ','.join(
                list(map(lambda x: self.extract_mapping_value_rec(x, main_table, version), v_array))) + ')'
        else:
            val = v_array[0]
            if self.is_simple_ref(val, main_table) == True:
                res['source'] = normalize_field_name(val['field'])
            else:
                v = self.extract_mapping_value_rec(val, main_table, version)
                if v.startswith('aux.'):
                    res['function'] = v
                else:
                    res['value'] = v

    def extract_mapping_value_rec(self, val, main_table, version):
        type_ = type(val)
        # should be direct value
        if type_ is dict and 'apply_function' not in val:
            if version == 'v1':
                if val['table'] == main_table:
                    return '_source.' + normalize_field_name(val['field'])
                else:
                    if 'aggregator' in val:
                        return 'len(_source.' + val['table'] + ')'
                    else:
                        return '_source.' + val['table'] + '.' + normalize_field_name(val['field'])
            else:
                return normalize_field_name(val['field'])
        # function application
        if type_ is dict and 'apply_function' in val:
            # single parameter
            if type(val['to']) is dict:
                return 'aux.' + self.to_func_name(val['apply_function']) + '(' + self.extract_mapping_value_rec(
                    val['to'],
                    main_table,
                    version) + ')'
            # array
            else:
                fvarr = []
                for fv in val['to']:
                    fvarr.append(self.extract_mapping_value_rec(fv, main_table, version))
                return 'aux.' + self.to_func_name(val['apply_function']) + '(' + ', '.join(fvarr) + ')'
        # literal
        else:
            return "'" + val + "'"

    # convert function name
    def to_func_name(self, str):
        return normalize_field_name(str)

    # return false if the value is a function call
    def is_simple_ref(self, val, main_table):
        type_ = type(val)
        if type_ is dict and 'apply_function' not in val:
            # return True if 'table' in val and val['table'] == main_table else False
            return True if 'table' in val else False
        return False

    def create_explicit_pk_mapping(self, sk, tk):
        if len(sk) != len(tk):
            self.logging.log_generator_error('INCORRECT MAPPING: key mapping not aligned', False)
            raise Exception('INCORRECT MAPPING: key mapping not aligned')
        res = [
            {'field': normalize_field_name(t['field']), 'source': normalize_field_name(s['field'])}
            for s, t in zip(sk, tk)
            ]
        return res

    def add_multicondition(self, mc, m_out, s_spec):
        (pre_tables, exp) = self.process_condition(mc, s_spec)
        indices = set()
        for pt in pre_tables: indices.add(pt['table'] + (pt['alias'] if 'alias' in pt else ''))
        maxRel = None
        for i in range(len(m_out['relations'])):
            rel = m_out['relations'][i]
            n = rel['relation'] + (rel['alias'] if 'alias' in rel else '')
            if n in indices: maxRel = rel
        if maxRel != None:
            if 'multicondition' not in maxRel: maxRel['multicondition'] = []
        maxRel['multicondition'].append(exp)

    # create mappings for Primary Key. In a simple case consider one-to-one mapping for all the key components (same types and cardinality)
    # In more complex case the PK mapping definition should be explictly provided
    def create_pk_mapping(self, main_def, main_target_def, config_spec):
        res = []
        # find explicit mapping function
        m_func = None
        if 'pk_mappings' in config_spec:
            for m in config_spec['pk_mappings']:
                if m['from'] == main_def['name'] and m['to'] == main_target_def['name']:
                    m_func = m
                    break

        if m_func != None:
            pkey = {}
            pkey['field'] = main_target_def['schema']['primaryKey']
            fvarr = []
            for fv in m['params']:
                fvarr.append(self.extract_mapping_value_rec(fv, main_def['name']))
            pkey['value'] = 'aux.' + self.to_func_name(m['function']) + '(' + ', '.join(fvarr) + ')'
            res.append(pkey)
            return res

        # add primary key: assume same type and cardinality
        if 'primaryKey' in main_def['schema'] and 'primaryKey' in main_target_def['schema']:
            if type(main_def['schema']['primaryKey']) is str:
                pkey = {}
                pkey['field'] = main_target_def['schema']['primaryKey']
                pkey['source'] = main_def['schema']['primaryKey']
                res.append(pkey)
            else:
                for i in range(len(main_def['schema']['primaryKey'])):
                    pkey = {}
                    pkey['field'] = main_target_def['schema']['primaryKey'][i]
                    pkey['source'] = main_def['schema']['primaryKey'][i]
                    res.append(pkey)
        return res

    # create a single migration definition
    def create_migration_def(self, mapping, s_spec, t_spec, c_spec, version):
        m_out = {}

        # main table. Extract name and alias
        main_table = self.extract_source(mapping, s_spec)
        main_table_def = self.get_table_name_rev(main_table)
        m_out['source'] = main_table_def[0]
        if len(main_table_def) == 2: m_out['alias'] = main_table_def[1]

        m_out['target'] = self.extract_target(mapping['target_tables'], t_spec)
        m_out['selectionCriteria'] = []
        rel_conditions = {}
        rel_aggs = {}
        rel_fields = {}

        m_out['relations'] = []

        # process table diff scenario: source tables with 'not_' prefix.
        # currently support no other conditions / field defs on this table
        for t in mapping['source_tables']:
            if t.startswith('not_'):
                original_name = t[4:]
                if version == 'v1':
                    fields = []
                    for spec in s_spec['resources']:
                        if spec['name'] == original_name:
                            fields = [spec['schema']['primaryKey']] if type(spec['schema']['primaryKey']) is str else \
                                spec['schema']['primaryKey']
                    rel = {'join': 'left_only', 'fields': fields, 'relation': original_name}
                    m_out['selectionCriteria'].append('len(_source.' + original_name + ') == 0')
                else:
                    rel = {'join': 'left_only', 'fields': [], 'relation': original_name}
                m_out['relations'].append(rel)

        # process pre-conditions to prepare single table condition list or selection criteria
        multitable_conditions = []
        cond_fields = {}
        for pre in mapping['pre_conditions']:
            unique_tables = []
            (pre_tables, exp) = self.process_condition(pre, s_spec)
            if exp is None: continue
            for pt in pre_tables:
                unique_tables.append(pt['table'])
                if pt['table'] in cond_fields:
                    cond_fields[pt['table']].append(pt['field'])
                else:
                    cond_fields[pt['table']] = [pt['field']]
            pre_tables = list(dict.fromkeys(unique_tables))
            if len(pre_tables) == 1 and pre_tables[0] == main_table:
                m_out['selectionCriteria'].append(exp)
            elif len(pre_tables) == 1:
                cond_list = self.get_or_create_conditions_for_rel(pre_tables[0], rel_conditions, s_spec)
                cond_list.append(exp)
            elif 'key' not in pre:
                # TODO
                multitable_conditions.append(pre)
                # raise Exception('NON SUPPORTED MAPPING: multi-table condition: ' + str(pre))

        if len(m_out['selectionCriteria']) > 0:
            m_out['selectionCriteria'] = ' and '.join(m_out['selectionCriteria'])

        # process source fields to populate relations definition
        if type(mapping['source_fields']) is dict:
            mapping['source_fields'] = [mapping['source_fields']]
        for tc in cond_fields:
            found = False
            for ts in mapping['source_fields']:
                if ts['table'] == tc:
                    found = True
                    break
            if found == False:
                mapping['source_fields'].append({'table': tc, 'fields': cond_fields[tc]})
        # reset map of fields
        cond_fields = {}
        for field in mapping['source_fields']:
            if field['table'] not in cond_fields:
                cond_fields[field['table']] = field['fields']

        unique_relations = set()
        for j in mapping['joins']:
            jt = j['join'][1]
            rel = {}
            tn_def = self.get_table_name_rev(jt['table'])
            rel['relation'] = tn_def[0]
            if len(tn_def) > 1: rel['alias'] = tn_def[1]
            unique_rel_name = rel['relation'] + (rel['alias'] if 'alias' in rel else '')
            if unique_rel_name in unique_relations: continue
            unique_relations.add(unique_rel_name)

            rel['join'] = 'both'
            rel['fields'] = []
            rel['source_fields'] = [normalize_field_name(j['join'][0]['field'])]
            rel['target_fields'] = [normalize_field_name(jt['field'])]
            aggs = {}
            fields = cond_fields[jt['table']] if jt['table'] in cond_fields else []
            rel_fields[jt['table']] = {'table': jt['table'], 'fields': fields}
            for field_def in fields:
                if type(field_def) is str:
                    fld_normalized = normalize_field_name(field_def)
                    if fld_normalized not in rel['fields']:
                        rel['fields'].append(normalize_field_name(field_def))
                else:
                    agg_f_name = normalize_field_name(field_def['field'])
                    if agg_f_name not in rel['fields']:
                        rel['fields'].append(agg_f_name)
                    aggs[agg_f_name] = field_def['aggregator']
            if jt['table'] in rel_conditions:
                rel['condition'] = rel_conditions[jt['table']]
            if len(aggs) > 0:
                rel['aggregations'] = aggs
                rel_aggs[jt['table']] = aggs
            m_out['relations'].append(rel)

        for field in mapping['source_fields']:
            if field['table'] in cond_fields:
                field['fields'] += cond_fields[field['table']]
            if field['table'] == main_table:
                continue
            if field['table'] in rel_fields:
                continue
            rel_fields[field['table']] = field
            rel = {'relation': field['table'], 'join': 'both', 'fields': []}
            aggs = {}
            for field_def in field['fields']:
                if type(field_def) is str:
                    fld_normalized = normalize_field_name(field_def)
                    if fld_normalized not in rel['fields']:
                        rel['fields'].append(normalize_field_name(field_def))
                else:
                    # aggs.append({'field': normalize_field_name(field_def['field']), 'aggregator': field_def['aggregator']})
                    agg_f_name = normalize_field_name(field_def['field'])
                    if agg_f_name not in rel['fields']:
                        rel['fields'].append(agg_f_name)
                    aggs[agg_f_name] = field_def['aggregator']
            if field['table'] in rel_conditions:
                rel['condition'] = rel_conditions[field['table']]
            if len(aggs) > 0:
                rel['aggregations'] = aggs
                rel_aggs[field['table']] = aggs
            m_out['relations'].append(rel)

        for mc in multitable_conditions:
            self.add_multicondition(mc, m_out, s_spec)

        # check there are non undeclared fields used
        for key in rel_conditions:
            if key not in rel_fields:
                self.logging.log_generator_error(
                    'INCORRECT MAPPING: condition fields are not listed as source fields: {}'.format(key), False)
                raise Exception('INCORRECT MAPPING: condition fields are not listed as source fields: {}'.format(key))

        m_out['fieldMappings'] = []

        main_def = next(t for t in s_spec['resources'] if t['name'] == m_out['source'])
        main_target_def = next(t for t in t_spec['resources'] if t['name'] == m_out['target'])

        # add primary key mappings
        pk_mappings = self.create_explicit_pk_mapping(mapping['source_key'],
                                                      mapping[
                                                          'target_key']) if 'source_key' in mapping else self.create_pk_mapping(
            main_def,
            main_target_def,
            c_spec)
        if len(pk_mappings) > 0:
            m_out['fieldMappings'].extend(pk_mappings)

        for rule in mapping['rules']:
            r_out = {}
            r_out['field'] = normalize_field_name(rule['field']['field'])
            self.extract_mapping_value(rule['value'], r_out, main_table, version)
            main_alias = m_out['alias'] if 'alias' in m_out else ''
            alias = map(lambda v: v['as'], filter(lambda v: 'as' in v and v['as'] != main_alias, rule['value']))
            alias = list(dict.fromkeys(alias))
            if len(alias) > 0: r_out['alias'] = alias[0]

            m_out['fieldMappings'].append(r_out)

        return m_out

    def spec_compatible(self, m1, m2):
        if dict.fromkeys(m1['source_tables']) != dict.fromkeys(m2['source_tables']) or dict.fromkeys(
                m1['target_tables']) != dict.fromkeys(m2['target_tables']): return False

        s1 = set(map(lambda r: r['comment'], m1['pre_conditions']) if 'pre_conditions' in m1 else [])
        s2 = set(map(lambda r: r['comment'], m2['pre_conditions']) if 'pre_conditions' in m2 else [])

        j1 = set(map(lambda r: r['comment'], m1['joins']) if 'joins' in m1 else [])
        j2 = set(map(lambda r: r['comment'], m2['joins']) if 'joins' in m2 else [])

        if len(s1) != len(s2): return False
        if len(j1) != len(j2): return False

        if not s1.issubset(s2) and not s2.issubset(s1): return False
        if not j1.issubset(j2) and not j2.issubset(j1): return False

        return True

    def merge_spec(self, m1, m2):
        pc1 = m1['post_conditions'] if 'post_conditions' in m1 else []
        pc2 = m2['post_conditions'] if 'post_conditions' in m2 else []
        m1['post_conditions'] = pc1 + pc2

        if type(m1['source_fields']) is dict:
            m1['source_fields'] = [m1['source_fields']]
        if type(m2['source_fields']) is dict:
            m2['source_fields'] = [m2['source_fields']]
        d = {}
        for sf in m1['source_fields']:
            d[sf['table']] = sf['fields']
        for sf in m2['source_fields']:
            if sf['table'] not in d:
                d[sf['table']] = sf['fields']
            else:
                d[sf['table']] += sf['fields']
        m1['source_fields'] = []
        for t in d:
            m1['source_fields'].append({'table': t, 'fields': list(dict.fromkeys(d[t]))})

        if type(m1['target_fields']) is dict:
            m1['target_fields'] = [m1['target_fields']]
        if type(m2['target_fields']) is dict:
            m2['target_fields'] = [m2['target_fields']]
        d = {}
        for sf in m1['target_fields']:
            d[sf['table']] = sf['fields']
        for sf in m2['target_fields']:
            if sf['table'] not in d:
                d[sf['table']] = sf['fields']
            else:
                d[sf['table']] += sf['fields']
        m1['target_fields'] = []
        for t in d:
            m1['target_fields'].append({'table': t, 'fields': d[t]})

        for r in m2['rules']:
            m1['rules'].append(r)

        return m1

    def preprocess_rec(self, arg):
        t = type(arg)
        res = []
        if t is list:
            for e in arg:
                res = res + self.preprocess_rec(e)
        elif t is dict and 'table' in arg:
            arg['table'] = self.get_table_name(arg)
            return [{'table': arg['table'], 'field': arg['field']}]
        elif t is dict:
            if 'arg0' in arg:
                res = res + self.preprocess_rec(arg['arg0'])
            if 'arg1' in arg:
                res = res + self.preprocess_rec(arg['arg1'])
        return res

    def preprocess_condition(self, pre):
        if 'key' in pre:
            for k in pre:
                k['table'] = self.get_table_name(k)
            return map(lambda k: {'table': k['table'], 'field': k['field']}, pre['key'])
        pre_tables = self.preprocess_rec(pre['arg0'])
        if 'arg1' in pre:
            pre_tables += self.preprocess_rec(pre['arg1'])
        return pre_tables

    def norm_condition(self, pre):
        if 'arg0' in pre:
            if 'column' in pre['arg0']: pre['arg0'] = pre['arg0']['column'][0]
            self.norm_condition(pre['arg0'])
        if 'arg1' in pre:
            if 'column' in pre['arg1']: pre['arg1'] = pre['arg1']['column'][0]
            self.norm_condition(pre['arg1'])

    def preprocess_mapping(self, m):
        # ignore for the moment
        m['source_tables'] = []
        m['target_tables'] = []
        m['source_fields'] = []
        m['target_fields'] = []

        table_positions = {}
        # normalize
        joins = []
        pre_conds = []
        for r in m['rules']:
            r['field'] = r['field'][0]

        m['pre_conditions'].sort(key=lambda p: self.get_table_name(p['key'][0]) if 'key' in p else '_')
        for p in m['pre_conditions']:
            if 'key' in p:
                if 'source_key' not in m:
                    for el in p['key']:
                        el['table'] = self.get_table_name(el)
                        table_positions[el['table']] = len(table_positions)
                    m['source_key'] = p['key']
                continue
            if 'join' in p:
                arr = []
                for jelem in p['join']:
                    if isinstance(jelem, list):
                        arr.append(jelem[0])
                    else:
                        arr.append(jelem)
                p['join'] = arr
                joins.append(p)
                continue
            self.norm_condition(p)
            pre_conds.append(p)
        for p in m['post_conditions']:
            if 'key' in p:
                m['target_key'] = p['key']
                continue
        m['pre_conditions'] = pre_conds
        m['joins'] = joins

        cond_fields = {}
        unique_tables = set()
        for pre in m['pre_conditions']:
            pre_tables = self.preprocess_condition(pre)
            for pt in pre_tables:
                unique_tables.add(pt['table'])
                if pt['table'] in cond_fields:
                    cond_fields[pt['table']].append(pt['field'])
                else:
                    cond_fields[pt['table']] = [pt['field']]

        for r in m['rules']:
            m['target_tables'].append(r['field']['table'])
            m['target_fields'].append({'table': r['field']['table'], 'fields': [r['field']['field']]})
            for v in r['value']:
                if 'table' in v:
                    tn = self.get_table_name(v)
                    v['table'] = tn
                    unique_tables.add(tn)
                    if tn in cond_fields:
                        cond_fields[tn].append(v['field'])
                    else:
                        cond_fields[tn] = [v['field']]

        for r in m['joins']:
            for v in r['join']:
                if 'table' in v:
                    tn = self.get_table_name(v)
                    v['table'] = tn
                    unique_tables.add(tn)
                    if tn in cond_fields:
                        cond_fields[tn].append(v['field'])
                    else:
                        cond_fields[tn] = [v['field']]

        # process source fields to populate relations definition
        for tc in cond_fields:
            fields = list(dict.fromkeys(cond_fields[tc]))
            m['source_fields'].append({'table': tc, 'fields': fields})
        # for f in m['source_fields']:
        #     if f['table'] in cond_fields:
        #         f['fields'] += cond_fields[f['table']]

        m['source_tables'] = list(dict.fromkeys(unique_tables))
        m['target_tables'] = list(dict.fromkeys(m['target_tables']))

        res_joins = []
        l = len(m['joins'])

        if len(table_positions) == 0:
            self.logging.log_generator_error('INVALID MAPPING: key definition is missing', False)
            raise Exception('INVALID MAPPING: key definition is missing')

        while len(res_joins) < l:
            for pos in range(len(m['joins'])):
                j = m['joins'][pos]
                found = False
                for elem in j['join']:
                    if elem['table'] in table_positions:
                        found = True
                        break
                # print(found)
                if found:
                    res_joins.append(j)
                    for elem in j['join']:
                        if elem['table'] not in table_positions:
                            table_positions[elem['table']] = len(table_positions)
                    del m['joins'][pos]
                    break

        for j in res_joins:
            j['join'].sort(key=lambda el: table_positions[el['table']])
        m['joins'] = res_joins

    # Transform m1 in a new specification, removing common mapping rules with m2. If there is nothing to remove,
    # return None
    def reduce(self, m1, m2):
        r1 = m1['rules'] if 'rules' in m1 else []
        r2 = set(map(lambda r: r['comment'], m2['rules']) if 'rules' in m2 else [])
        r = []
        for elem in r1:
            if elem['comment'] not in r2:
                r.append(elem)
        if len(r) == len(r1): return None
        m1['rules'] = r
        return m1

    # Compare whether one specification is more generic than other: speak about the same tables and relations, but
    # conditions are less restrictive
    def more_generic(self, m1, m2):
        if dict.fromkeys(m1['source_tables']) != dict.fromkeys(m2['source_tables']) or dict.fromkeys(
                m1['target_tables']) != dict.fromkeys(m2['target_tables']): return None
        j1 = set(map(lambda r: r['comment'], m1['joins']) if 'joins' in m1 else [])
        j2 = set(map(lambda r: r['comment'], m2['joins']) if 'joins' in m2 else [])
        if len(j1) != len(j2): return None
        if not j1.issubset(j2) or not j2.issubset(j1): return None

        s1 = set(map(lambda r: r['comment'], m1['pre_conditions']) if 'pre_conditions' in m1 else [])
        s2 = set(map(lambda r: r['comment'], m2['pre_conditions']) if 'pre_conditions' in m2 else [])

        if s1.issubset(s2) and len(s1) != len(s2): return 1
        if s2.issubset(s1) and len(s1) != len(s2): return -1
        if s1.issubset(s2) and s2.issubset(s1): return 0

        return None

    def more_generic2(self, m1, m2):
        if dict.fromkeys(m1['source_tables']) != dict.fromkeys(m2['source_tables']) or dict.fromkeys(
                m1['target_tables']) != dict.fromkeys(m2['target_tables']): return None
        j1 = set(map(lambda r: r['comment'], m1['joins']) if 'joins' in m1 else [])
        j2 = set(map(lambda r: r['comment'], m2['joins']) if 'joins' in m2 else [])
        if not j1.issubset(j2) and not j2.issubset(j1): return None

        s1 = set(map(lambda r: r['comment'], m1['pre_conditions']) if 'pre_conditions' in m1 else [])
        s2 = set(map(lambda r: r['comment'], m2['pre_conditions']) if 'pre_conditions' in m2 else [])

        if s1.issubset(s2) and len(s1) != len(s2): return 1 if j1.issubset(j2) else None
        if s2.issubset(s1) and len(s1) != len(s2): return -1 if j2.issubset(j1) else None
        if s1.issubset(s2) and s2.issubset(s1): return 0 if j1.issubset(j2) and j2.issubset(j1) else None

        return None

    # Check if the two specification have the same set of rules
    def same_rules(self, m1, m2):
        if dict.fromkeys(m1['source_tables']) != dict.fromkeys(m2['source_tables']) or dict.fromkeys(
                m1['target_tables']) != dict.fromkeys(m2['target_tables']): return None
        j1 = set(map(lambda r: r['comment'], m1['joins']) if 'joins' in m1 else [])
        j2 = set(map(lambda r: r['comment'], m2['joins']) if 'joins' in m2 else [])
        if len(j1) != len(j2): return None
        if not j1.issubset(j2) or not j2.issubset(j1): return None

        r1 = set(map(lambda r: r['comment'], m1['rules']) if 'rules' in m1 else [])
        r2 = set(map(lambda r: r['comment'], m2['rules']) if 'rules' in m2 else [])
        return r1.issubset(r2) and r2.issubset(r1)

    # Update the specification adding another set of pre_conditions
    def union_conditions(self, m1, m2, s_spec):
        # group conditions by table. Check that the conditions differ for at most 1 table. Otherwise merge is not possible.
        # transform conditions for that table into 'OR' form.
        table_conds1 = {}
        table_conds2 = {}
        table_conds_full1 = {}
        table_conds_full2 = {}
        c1 = m1['pre_conditions'] if 'pre_conditions' in m1 else []
        c2 = m2['pre_conditions'] if 'pre_conditions' in m2 else []
        for pre in c1:
            (pre_tables, exp) = self.process_condition(pre, s_spec)
            if exp is None: continue
            for pt in pre_tables:
                tname = pt['table']
                if tname not in table_conds1:
                    table_conds1[tname] = set()
                    table_conds_full1[tname] = []
                if pre['comment'] not in table_conds1[tname]:
                    table_conds1[tname].add(pre['comment'])
                    table_conds_full1[tname].append(pre)
        for pre in c2:
            (pre_tables, exp) = self.process_condition(pre, s_spec)
            if exp is None: continue
            for pt in pre_tables:
                tname = pt['table']
                if tname not in table_conds2:
                    table_conds2[tname] = set()
                    table_conds_full2[tname] = []
                if pre['comment'] not in table_conds2[tname]:
                    table_conds2[tname].add(pre['comment'])
                    table_conds_full2[tname].append(pre)
        if abs(len(table_conds1) - len(table_conds2)) > 1:
            return None
        table = None
        for t in table_conds1:
            if t not in table_conds2 or table_conds1[t] != table_conds2[t]:
                if table is not None:
                    return None
                table = t
        for t in table_conds2:
            if t not in table_conds1 and table is not None:
                return None
            if t not in table_conds1: table = t

        if table is None:
            table_diff = set(table_conds1) - set(table_conds2) if len(table_conds1) > len(table_conds2) else set(
                table_conds2) - set(table_conds1)
            if len(table_diff) == 0:
                # print('Found the same')
                return m1
            table = list(table_diff)[0]
        new_pres = []
        for pre in table_conds_full1:
            if table != pre or table not in table_conds_full2:
                new_pres.append(self.to_op(table_conds_full1[pre], 'and'))
            else:
                con1 = self.to_op(table_conds_full1[pre], 'and')
                con2 = self.to_op(table_conds_full2[pre], 'and')
                new_pres.append(self.to_op([con1, con2], 'or'))

        if table not in table_conds_full1:
            new_pres.append(self.to_op(table_conds_full2[table], 'and'))

        m1['pre_conditions'] = new_pres
        # print('Found different')
        return m1

    def to_op(self, conds, op):
        if len(conds) == 1: return conds[0]
        arg0 = conds[0]
        arg1 = self.to_op(conds[1:], op)
        return {
            'arg0': arg0, 'arg1': arg1, 'operator': op,
            'comment': op + '(' + arg0['comment'] + ',' + arg1['comment'] + ')'
            }

    # optimize number of specifications. Split and join different mapping:
    # Let s = <C,M> is a mapping spec with C is a set of conditions (conjunction of) and M is a set of mapping rules.
    # Let '>' mean set inclusion, '*' mean set intersection, '+' union (or disjunction), '-' set difference, '0' empty set.
    #
    # 1. Given s1=<C1,M1> and s2=<C2,M2>, if C1 < C2 and M1*M2 != 0, then we can transform s2 in mapping s2' so that s2'=<C2, M2-M1>
    #
    # 2. Given s1=<C1,M> and s2=<C2,M>, we can merge s1 and s2 to s' = <C1 + C2, M>
    def optimize_spec(self, list, s_spec):
        list_new = list.copy()
        # implementing rule 1
        list_new = self.optimize_spec_rule1(list_new, s_spec)
        # implementing rule 2
        while True:
            changed = False
            for i1 in range(len(list_new)):
                s1 = list_new[i1]
                for i2 in range(i1 + 1, len(list_new)):
                    s2 = list_new[i2]
                    if self.same_rules(s1, s2):
                        # print('same rules: ' + str(i1) +' and '+ str(i2))
                        if self.union_conditions(s1, s2, s_spec) is not None:
                            list_new.pop(i2)
                            changed = True
                            break
                if changed: break
            if not changed:
                break

        return list_new

    def optimize_spec_rule1(self, list_new, s_spec):
        while True:
            changed = False
            for i1 in range(len(list_new)):
                s1 = list_new[i1]
                for i2 in range(i1 + 1, len(list_new)):
                    s2 = list_new[i2]
                    comp = self.more_generic2(s1, s2)
                    # comp2 = more_generic2(s1, s2)
                    # if comp != comp2:
                    #     print('Difference comparing ' + str(i1) +' and '+ str(i2))
                    if comp == 0:
                        list_new.pop(i2)
                        changed = True
                        break
                    if comp == 1:
                        r = self.reduce(s2, s1)
                        if r != None:
                            if len(r['rules']) == 0:
                                list_new.pop(i2)
                            else:
                                list_new[i2] = r
                            changed = True
                            break
                    if comp == -1:
                        r = self.reduce(s1, s2)
                        if r != None:
                            if len(r['rules']) == 0:
                                list_new.pop(i1)
                            else:
                                list_new[i1] = r
                            changed = True
                            break
                if changed: break
            if not changed:
                break
        return list_new

    def check_compatibility(self, specifications_list, s_spec):
        res_list = []
        i = 0
        while i < len(specifications_list):
            j = i + 1
            m1 = specifications_list[i]
            while j < len(specifications_list):
                m2 = specifications_list[j]
                if self.spec_compatible(m1, m2):
                    m1 = self.merge_spec(m1, m2)
                    del specifications_list[j]
                else:
                    j += 1
            res_list.append(m1)
            i += 1
        res_list = self.optimize_spec(res_list, s_spec)
        return res_list

    def read_dataset_config(self, config, folder, source_datapkg, target_datapkg):
        output = os.path.join(config.domain["domain_settings_path"], folder,
                              "migration_" + folder + '_result.json')
        version = config.converter["version"]
        c_spec = {}
        with open(source_datapkg) as f:
            s_spec = json.load(f)
        with open(target_datapkg) as f:
            t_spec = json.load(f)
        input_file_to_convert = os.path.join(config.domain["domain_settings_path"], folder,
                                             "{source}_to_{target}_{folder}.json".format(source=s_spec['name'],
                                                                                         target=t_spec['name'],
                                                                                         folder=folder))
        with open(input_file_to_convert) as f:
            spec = json.load(f)
        if 'config' in config.converter:
            cfile = config.converter['config']
            with open(cfile) as f: c_spec = json.load(f)
        return c_spec, output, s_spec, spec, t_spec, version

    def run_converter(self):
        self.logging.log_generator("Starting Converter...", False)
        source_datapkg = self.config.domain["domain_settings_path"] + self.config.domain['package_source'] + '.json'
        target_datapkg = self.config.domain["domain_settings_path"] + self.config.domain['package_target'] + '.json'
        datasets_folders = os.listdir(self.config.domain["domain_settings_path"])
        for folder in datasets_folders:
            if os.path.isdir(os.path.join(self.config.domain["domain_settings_path"], folder)):
                c_spec, output, s_spec, spec, t_spec, version = self.read_dataset_config(self.config, folder, source_datapkg,
                                                                                         target_datapkg)
                output_spec = {
                    'name': s_spec['name'] + '2' + t_spec['name'], 'description': '', 'sourcePackage': s_spec['name'],
                    'targetPackage': t_spec['name'], 'mappings': []
                    }
                if 'functions' in c_spec:
                    output_spec['imports'] = {os.path.basename(c_spec['functions']): c_spec['functions']}
                else:
                    output_spec['imports'] = {"aux": "aux_functions.py"}
                # process single mapping
                unique_set = set()
                specifications_list = list.copy(spec)
                for mapping in specifications_list:
                    self.preprocess_mapping(mapping)

                result_list = self.check_compatibility(specifications_list, s_spec)
                spec = result_list

                for num, mapping in enumerate(spec):
                    m_def = self.create_migration_def(mapping, s_spec, t_spec, c_spec, version)
                    str_def = str(m_def)
                    if str_def not in unique_set:
                        unique_set.add(str_def)
                        m_def['mapping_index'] = num + 1
                        output_spec['mappings'].append(m_def)

                with open(output, 'w') as outfile:
                    json.dump(output_spec, outfile, indent=4)
        self.logging.log_generator("Ending converter...")


if __name__ == "__main__":
    configs = configurations.Config("examples/SIB2SICRA/convert.ini")
    converter = Converter(configs)
    converter.run_converter()
