import migration_machine.config as configurations
import multiprocessing as mp
import xml.etree.ElementTree as ET
import clingo
import json
from collections import defaultdict
import pprint
import os

pp = pprint.PrettyPrinter(width=80, compact=True)

WORKING_DIR = "/tmp/migration_machine/"
_CONJ_CND_COUNTER = 0
_UNKN_CND_COUNTER = 0
SAVE_LP = True
SAVE_MODELS = True
SAVE_JSON = True

if not os.path.exists(WORKING_DIR):
    os.mkdir(WORKING_DIR)


class Resource:

    def __init__(self, source_or_target, name=None):
        self.source_or_target = source_or_target
        self.tables_onto_maps = dict()
        self.columns_onto_maps = dict()
        self.name = name

    def load(self, xml_file):
        resource_xml = ET.parse(xml_file).getroot()
        for table_xml in resource_xml.findall("tablemap"):
            self.add_table_onto_map(table_xml)
        for column_xml in resource_xml.findall('columnsmap'):
            self.add_column_onto_map(column_xml)

    def add_table_onto_map(self, table_xml):
        table_name = table_xml.attrib['name']
        tom = Table_onto_map(table_name)
        for map_to_class in table_xml.findall("class"):
            tom.add_maps_to_classes(map_to_class)
        if table_name not in self.tables_onto_maps:
            self.tables_onto_maps[table_name] = []
        self.tables_onto_maps[table_name].append(tom)

    def add_column_onto_map(self, column_xml):
        columns_name = tuple([column.attrib['name'] for column in column_xml.findall("column")])
        com = Column_onto_map(columns_name)
        for map_to_relation in column_xml.findall('relation'):
            com.add_maps_to_relations(map_to_relation)
        if columns_name not in self.columns_onto_maps:
            self.columns_onto_maps[columns_name] = []
        self.columns_onto_maps[columns_name].append(com)


class Table_onto_map():
    def __init__(self, name):
        self.name = name
        self.maps_to_classes = dict()
        self.class_defs = dict()

    def add_maps_to_classes(self, map_to_class_xml):
        class_names = map_to_class_xml.attrib['name'].split(",")
        condition_exprs = [_parse_condition(condition) for condition in map_to_class_xml.findall('condition')]
        if len(class_names) > 1:
            class_name = "_and_".join(class_names).replace("-", "not_")
            self.class_defs[class_name] = class_names
        else:
            class_name = class_names[0]
        if class_name not in self.maps_to_classes:
            self.maps_to_classes[class_name] = dict()
        self.maps_to_classes[class_name]["conditions"] = condition_exprs

    def to_lp(self, source_or_target):
        assert source_or_target in ['source', 'target']
        lp = f"%%%%%% BEGIN MAPPING {source_or_target} TABLE {self.name} " \
             f"TO {list(self.maps_to_classes.keys())} %%%%%%% \n\n"
        lp += '{}(table("{}")).\n'.format(source_or_target, self.name)
        conds, lp_conds = _generate_condition_clauses(self.maps_to_classes, source_or_target)
        lp += lp_conds
        for onto_class in self.maps_to_classes:
            if onto_class[0] == "-":
                not_onto_class = onto_class[1:]
            else:
                not_onto_class = "-" + onto_class
            if onto_class not in conds:
                lp += f'{onto_class} :- table("{self.name}").\n'
                lp += f'has(R,{onto_class}) :- has(R,table("{self.name}")).\n'
                lp += f'table("{self.name}") :- {onto_class}.\n'
                lp += f'has(R,table("{self.name}")) :- has(R,{onto_class}).\n'
                lp += f'-table("{self.name}") :- {not_onto_class}.\n'

            else:
                lp += f'{onto_class} :- table("{self.name}"),{conds[onto_class]}.\n'
                lp += f'has(R,{onto_class}) :- has(R,table("{self.name}")),has(R,{conds[onto_class]}).\n'
                lp += f'table("{self.name}") :- {onto_class}.\n'
                lp += f'has(R,table("{self.name}")) :- has(R,{onto_class}).\n'
                lp += f'{conds[onto_class]} :- table("{self.name}"), {onto_class}.\n'
                lp += f'has(R,{conds[onto_class]}) :- has(R,table("{self.name}")), has(R,{onto_class}).\n'
                lp += f'-table("{self.name}") :- {not_onto_class},{conds[onto_class]}.\n'

        # if source_or_target == "source" and conds.values():
        if conds.values():
            lp += f'0{"{"}{";".join(conds.values())}{"}"} 1 :- table("{self.name}").\n'
            scond_has_R_conds = [f"has(R,{cnd})" for cnd in conds.values()]
            lp += f'0{"{"}{";".join(scond_has_R_conds)}{"}"} 1 :- has(R,table("{self.name}")).\n'
        for complex_class in self.class_defs:
            lp += f"{complex_class} :- {', '.join(self.class_defs[complex_class])}.\n"
            body = ', '.join([f'has(R,{cl})' for cl in self.class_defs[complex_class]])
            lp += f"has(R,{complex_class}) :- {body}.\n"
            for basic_class in self.class_defs[complex_class]:
                lp += f"{basic_class} :- {complex_class}.\n"
                lp += f'has(R,{basic_class}) :- has(R,{complex_class}).\n'
        lp += f"\n\n%%%%%% END MAPPING {source_or_target} TABLE {self.name} " \
              f"TO {list(self.maps_to_classes.keys())} %%%%%%% \n\n"
        return (lp)


class Column_onto_map():
    def __init__(self, columns):
        self.name = columns  # the name can be a tuple of columns
        self.table = columns[0].split(".", maxsplit=1)[0]  # da sistemare nel caso ci siano colonne prese da piu tabella
        self.maps_to_classes = dict()
        self.maps_to_relations = dict()
        self.is_key = False

    def add_maps_to_relations(self, map_to_relation_xml):
        relation_name = map_to_relation_xml.attrib['name']
        condition_exprs = [_parse_condition(condition) for condition in map_to_relation_xml.findall('condition')]
        rel_map = {'conditions': condition_exprs}
        rel_map['domain'] = map_to_relation_xml.attrib['domain'] if 'domain' in map_to_relation_xml.attrib else None
        self.maps_to_relations[relation_name] = rel_map

    def to_lp(self, source_or_target):
        assert source_or_target in ['source', 'target']
        columns = self.name
        tables = set([col.split(".")[0] for col in columns])
        table_expr = "table({})".format(','.join(['"{}"'.format(table) for table in tables]))
        col_expr = f'column("{",".join(self.name)}")'
        lp = f"\n\n%%%%%% BEGIN MAPPING {source_or_target} COLUMN {self.name} " \
             f"TO {list(self.maps_to_relations.keys())} %%%%%%% \n\n"
        lp += f"{source_or_target}({col_expr}).\n"
        lp += f"{source_or_target}({table_expr}).\n"
        lp += f"hasColumn({table_expr},{col_expr}).\n"
        for relation, rel_map in self.maps_to_relations.items():
            conditions = rel_map['conditions']
            domain = rel_map['domain']
            body = [f'{source_or_target[0]}cond({cnd})' for cnd in conditions] + [table_expr]
            rBody = [f'{source_or_target[0]}cond(has(R,{cnd}))' for cnd in conditions] + [f'has(R,{table_expr})']
            if domain is not None:
                body += [domain]
                rBody += [f'has(R,{domain})']
            lp += f'has({relation},{col_expr}) :- {", ".join(body)}.\n'
            lp += f'has(R,has({relation},{col_expr})) :- {", ".join(rBody)}.\n'
        # if source_or_target == "source":
        conds, lp_conds = _generate_condition_clauses(self.maps_to_relations, source_or_target)
        if conds.values():
            lp += lp_conds
            lp += f'0{"{"}{";".join(conds.values())}{"}"} 1.\n'
            has_r_conds = [f'has(R,{cond})' for cond in conds.values()]
            lp += f'0{"{"}{";".join(has_r_conds)}{"}"} 1 :- has(R,table(_)).\n'
        lp += f"\n\n%%%%%% END MAPPING {source_or_target} COLUMN {self.name} " \
              f"TO {list(self.maps_to_relations.keys())} %%%%%%% \n\n"
        return lp


class Sourc_target_map():

    def __init__(self):
        self.source_tables = set()
        self.target_tables = set()
        self.source_columns = defaultdict(set)
        self.target_columns = defaultdict(set)
        self.rules = dict()
        self.pre_conditions = list()
        self.post_conditions = list()

    def serialize(self):
        serial = dict()
        serial["source_tables"] = list(self.source_tables)
        serial["target_tables"] = list(self.target_tables)
        serial["source_fields"] = list()
        for k, v in self.source_columns.items():
            kv = {"table": k, "fields": []}
            for field in v:
                if isinstance(field, tuple):
                    kv["fields"].append({
                        "aggregator": field[0],
                        "field": field[1]
                        })
                else:
                    kv["fields"].append(field)
                serial["source_fields"].append(kv)
        serial["target_fields"] = list()
        for k, v in self.target_columns.items():
            kv = {"table": k, "fields": []}
            for field in v:
                if isinstance(field, tuple):
                    kv["fields"].append({
                        "aggregator": field[0],
                        "field": field[1]
                        })
                else:
                    kv["fields"].append(field)
            serial["target_fields"].append(kv)
        serial["rules"] = [{
            "field": {
                "table": k[0],
                "field": k[1]
                },
            "value": v
            } for k, v in self.rules.items()]
        serial["pre_conditions"] = self.pre_conditions
        serial["post_conditions"] = self.post_conditions
        return serial

    def add_rule(self, atom):
        args = atom.arguments
        if args[0].name == "column":
            target_table, target_field = args[1].arguments[0].string.split(".", maxsplit=1)
            self.target_tables.add(target_table)
            self.target_columns[target_table].add(target_field)
            source_table, source_field = args[0].arguments[0].string.split(".", maxsplit=1)
            self.source_tables.add(source_table)
            self.source_columns[source_table].add(source_field)
            value = {"table": source_table, "field": source_field}
        elif args[0].name == "count":
            target_table, target_field = args[1].arguments[0].string.split(".", maxsplit=1)
            self.target_tables.add(target_table)
            self.target_columns[target_table].add(target_field)
            source_table, source_field = args[0].arguments[0].arguments[0].string.split(".", maxsplit=1)
            self.source_tables.add(source_table)
            value = {"table": source_table, "field": source_field, "aggregator": "count"}
            self.source_columns[source_table].add(("count", source_field))
        elif args[0].name == "eq":
            args = args[0].arguments
            target_table, target_field = args[0].arguments[0].string.split(".", maxsplit=1)
            self.target_tables.add(target_table)
            self.target_columns[target_table].add(target_field)
            value = args[1].string
        if len(args) == 3:
            value = {
                'apply_function': str(args[2]).replace("\"", ""),
                "to": {k: v for k, v in value.items()}
                }
        if (target_table, target_field) not in self.rules:
            self.rules[(target_table, target_field)] = list()
        self.rules[(target_table, target_field)].append(value)

    def add_condition(self, atom, pre_or_post):
        cond = dict()
        operator = atom.arguments[0].name
        cond["operator"] = operator
        if operator in ["eq", "neq", 'geq', "leq", "gt", 'lt']:
            for i in range(len(atom.arguments[0].arguments)):
                arg = atom.arguments[0].arguments[i]
                if arg.name == "column":
                    table, field = arg.arguments[0].string.split(".", maxsplit=1)
                    if pre_or_post == "pre":
                        self.source_tables.add(table)
                        self.source_columns[table].add(field)
                    if pre_or_post == "post":
                        self.target_tables.add(table)
                        self.target_columns[table].add(field)
                    cond['arg{}'.format(i)] = {"table": table, "field": field}
                else:
                    cond["arg{}".format(i)] = arg.string.replace("\"", "")
        elif operator == "oneOf":
            first_argument = atom.arguments[0].arguments[0]
            if first_argument.name == "column":
                table, field = first_argument.arguments[0].string.split(".", maxsplit=1)
                if pre_or_post == "pre":
                    self.source_tables.add(table)
                    self.source_columns[table].add(field)
                if pre_or_post == "post":
                    self.target_tables.add(table)
                    self.target_columns[table].add(field)
                cond["arg0"] = {"table": table, "field": field}
            cond['arg1'] = [val.string for val in atom.arguments[0].arguments[1:]]
        elif operator == "range":
            first_argument = atom.arguments[0].arguments[0]
            if first_argument.name == "column":
                table, field = first_argument.arguments[0].string.split(".", maxsplit=1)
                if pre_or_post == "pre":
                    self.source_tables.add(table)
                    self.source_columns[table].add(field)
                if pre_or_post == "post":
                    self.target_tables.add(table)
                    self.target_columns[table].add(field)
                cond["arg0"] = {"table": table, "field": field}
            cond['min'] = atom.arguments[0].arguments[1].string
            cond['max'] = atom.arguments[0].arguments[1].string
        if pre_or_post == "pre":
            self.pre_conditions.append(cond)
        elif pre_or_post == "post" and operator != "eq":
            self.post_conditions.append(cond)


class Source_target_list_of_maps(list):

    def __init__(self, source: Resource, target: Resource, ontology_paths, onto_rule_paths):
        super().__init__()
        self.source = source
        self.target = target
        self.base_lp = "\n\n".join([open(path, "r").read() for path in ontology_paths]) + open(onto_rule_paths,
                                                                                               "r").read()
        self.models = None
        self.st_tab_maps = dict()
        self.st_col_maps = dict()
        self.factorized_st_maps = dict()

    def map_tables(self, source_tables=None, target_tables=None):
        print(f'mapping {source_tables} to {target_tables}')
        models = []
        base_lp = self.base_lp
        for s_table, s_tableOntoMaps in self.source.tables_onto_maps.items():
            if source_tables is not None and s_table not in source_tables:
                continue
            s_tableOntoMaps_lp = "".join([tom.to_lp("source")
                                          for s_tableOntoMaps in self.source.tables_onto_maps.values()
                                          for tom in s_tableOntoMaps if tom.name != s_table and
                                          (source_tables is None or tom.name in source_tables)])
            s_keys_lp = ""
            for s_column, s_columnOntoMap in self.source.columns_onto_maps.items():
                for s_com in s_columnOntoMap:
                    if s_com.table == s_table and "haIdentificativo" in s_com.maps_to_relations:
                        s_keys_lp = s_com.to_lp("source")
            s_lp = s_tableOntoMaps_lp + f'\ntable("{s_table}").\n' + s_keys_lp
            # s_lp = f'\n table("{s_table}").\n'
            s_lp += "".join([s_tom.to_lp("source") for s_tom in s_tableOntoMaps])
            for t_table, t_tableOntoMaps in self.target.tables_onto_maps.items():
                cnt = -1
                if target_tables is not None and t_table not in target_tables:
                    continue
                for t_column, t_columnOntoMap in self.target.columns_onto_maps.items():
                    for t_com in t_columnOntoMap:
                        if t_com.table == t_table and "haIdentificativo" in t_com.maps_to_relations:
                            t_keys_lp = t_com.to_lp("target")
                for tom in t_tableOntoMaps:
                    cnt += 1
                    t_lp = tom.to_lp('target')
                    lp = base_lp + s_lp + t_lp + ":- not exists_tmap.\n" + t_keys_lp
                    st_table_map_dir = WORKING_DIR + t_table.replace('/', '-')
                    if not os.path.exists(st_table_map_dir):
                        os.mkdir(st_table_map_dir)
                    if SAVE_LP:
                        with open(f'{st_table_map_dir}/{s_table.replace("/", "-")}_{cnt}.lp', "w") as f:
                            print(f' writing {st_table_map_dir}/{s_table.replace("/", "-")}_{cnt}.lp')
                            f.write(lp)
                    new_models = compute_models(lp)
                    models += new_models
                    if SAVE_MODELS and len(new_models):
                        model_nr = 0
                        for m in new_models:
                            model_nr += 1
                            with open(f'{st_table_map_dir}/{s_table.replace("/", "-")}_{cnt}_{model_nr}.models',
                                      "w") as f:
                                for s in m:
                                    f.write(str(s) + "\n")
        self.st_tab_maps = compute_mappings_from(models)
        print("I have found the folloiwng mappings")
        print("\t" + "\n\t".join([str(a) for k in self.st_tab_maps for a in k]))
        print()

    def map_columns(self, source_tables=None, target_tables=None,
                    source_columns=None, target_columns=None):
        print(f'mapping the columns of {", ".join(source_tables)} to the columns of {", ".join(target_tables)}')
        if source_tables is None:
            source_tables = list(self.source.tables_onto_maps.keys())
        if target_tables is None:
            target_tables = list(self.target.tables_onto_maps.keys())
        if source_columns is None:
            source_columns = list(self.source.columns_onto_maps.keys())
        if target_columns is None:
            target_columns = list(self.target.columns_onto_maps.keys())
        models = []
        base_lp = self.base_lp
        s_tab_ontomap_lp = ""
        for s_table, s_tableOntoMaps in self.source.tables_onto_maps.items():
            if s_table in source_tables:
                s_tab_ontomap_lp += "".join([tom.to_lp("source") for tom in s_tableOntoMaps])
        for st_tab_map, st_tab_map_models in self.st_tab_maps.items():
            s_lp = ""
            for s_column, s_columnsOntoMaps in self.source.columns_onto_maps.items():
                if s_columnsOntoMaps[0].table not in source_tables:
                    continue
                s_lp += "".join([com.to_lp("source") for com in s_columnsOntoMaps])
                s_lp += s_tab_ontomap_lp
            count_st_tab_map_m = 0
            for m in st_tab_map_models:
                count_st_tab_map_m += 1
                st_lp = ".\n".join(str(a) for a in m['model']) + ".\n"
                # for m1 in st_tab_map_models:
                #     for scond in m1['core']['scond']:
                #         if scond not in m['core']['scond']:
                #             st_lp += f"-{scond}.\n"
                t_lp = ""
                for t_column, t_columnsOntoMaps in self.target.columns_onto_maps.items():
                    if t_columnsOntoMaps[0].table not in target_tables:
                        continue
                    if t_column not in target_columns:
                        continue
                    if str(st_tab_map[0].arguments[-1]).strip('"') == t_columnsOntoMaps[0].table:
                        t_lp += "".join([com.to_lp("target") for com in t_columnsOntoMaps])
                lp = base_lp + s_lp + t_lp + st_lp + ":- not exists_cmap.\n"
                st_tab_map_string = str(st_tab_map).replace("/", "_").replace("tmap", "").replace('"', '').replace(" ",
                                                                                                                   "")
                st_column_map_dir = f'{WORKING_DIR}/{st_tab_map_string}/'
                if not os.path.exists(st_column_map_dir):
                    os.mkdir(st_column_map_dir)
                if SAVE_LP:
                    with open(f"{st_column_map_dir}{count_st_tab_map_m}.lp", "w") as f:
                        f.write(lp)
                new_models = compute_models(lp)
                models += new_models
                if SAVE_MODELS and len(new_models):
                    model_nr = 0
                    for m in new_models:
                        model_nr += 1
                        with open(f"{st_column_map_dir}/{count_st_tab_map_m}_{model_nr}.models", "w") as f:
                            for s in m:
                                f.write(str(s) + "\n")
        self.st_col_maps = compute_mappings_from(models)
        print("I have found the folloiwng columns mappings")
        print("\t" + "\n\t".join([str(a) for k in self.st_col_maps for a in k if a.name == "cmap"]))
        print()

    def to_json(self):
        col_maps_json = []
        for maps, col_map in self.st_col_maps.items():
            for rule in col_map:
                m = rule['model']
                col_maps_with_conds_json = {"comment": ",".join([str(a) for a in get_mappings_from(m)['cmap']])}
                col_maps_with_conds_json["rules"] = []
                for map in get_mappings_from(m)["cmap"]:
                    col_map_json = {'comment': str(map)}
                    if len(map.arguments) == 4:
                        col_map_json['value'] = column_to_json(map.arguments[1:2], relation=map.arguments[0])
                    else:
                        col_map_json['value'] = column_to_json(map.arguments[:1])
                    if len(map.arguments) == 2:
                        col_map_json['field'] = column_to_json(map.arguments[-1:])
                    else:
                        col_map_json['field'] = column_to_json(map.arguments[-1:], relation=map.arguments[-2])
                    col_maps_with_conds_json["rules"].append(col_map_json)
                col_maps_with_conds_json['source_fields'] = [f for cm in col_maps_with_conds_json['rules'] for f in
                                                             cm['value']]
                col_maps_with_conds_json['target_fields'] = [f for cm in col_maps_with_conds_json['rules'] for f in
                                                             cm['field']]
                col_maps_with_conds_json['source_tables'] = list(
                    set([sf['table'] for sf in col_maps_with_conds_json['source_fields']]))
                col_maps_with_conds_json['target_tables'] = list(
                    set([sf['table'] for sf in col_maps_with_conds_json['target_fields']]))
                col_maps_with_conds_json["pre_conditions"] = list(
                    filter(None, [cond_to_json(a.arguments[0]) for a in
                                  get_conds_from(m)['scond']]))
                col_maps_with_conds_json["post_conditions"] = list(
                    filter(None, [cond_to_json(a.arguments[0]) for a in
                                  get_conds_from(m)['tcond']]))
                col_maps_json.append(col_maps_with_conds_json)
        return col_maps_json


class Mappings:
    def __init__(self, config):
        self.mappings_dir = None
        self.source2target = None
        self.config = config

    def read_configs(self):
        sib2000 = Resource("source", self.config.domain["package_source"])
        sicra = Resource("target", self.config.domain["package_target"])
        sib2000.load(self.config.modelling["domain_model_source_file"])
        sicra.load(self.config.modelling["domain_model_target_file"])
        ontology_paths = [self.config.modelling["domain_model_ontology_file"]]
        onto_rule_paths = self.config.modelling["domain_model_rules_file"]
        self.source2target = Source_target_list_of_maps(sib2000, sicra, ontology_paths, onto_rule_paths)
        self.mappings_dir = self.config.domain["domain_settings_path"]

    def map_correspondences(self, table):
        self.source2target.map_tables(source_tables=table[:-1], target_tables=table[-1:])
        self.source2target.map_columns(source_tables=table[:-1], target_tables=table[-1:])
        maps = self.source2target.to_json()
        folder = os.path.join(self.mappings_dir, table[-1])
        os.makedirs(folder, exist_ok=True)
        file_path = os.path.join(folder,
                                 f"{self.source2target.source.name}_to_{self.source2target.target.name}_{table[-1]}.json")
        with open(file_path, "w") as fp:
            json.dump(maps, fp, indent=3)


def cond_to_json(cond, relation=None):
    cond_json = dict()
    cond_json["comment"] = str(cond)
    if cond.name == "value":
        return str(cond.arguments[0]).strip('"')
    if cond.name == "table":
        if relation is None:
            return {cond.name: str(cond.arguments[0]).strip('"')}
        else:
            return {cond.name: str(cond.arguments[0]).strip('"') + "_" + str(relation)}
    elif cond.name == "column":
        return {cond.name: column_to_json(cond.arguments, relation=relation)}
    elif cond.name in ["in", 'and']:
        return
    elif cond.name == "key":
        result = {cond.name: key_to_json(cond.arguments)}
        result['comment'] = str(cond)
        return result
    elif cond.name == "join":
        result = {cond.name: join_to_json(cond.arguments)}
        result['comment'] = str(cond)
        return result
    elif cond.name == "has":
        return cond_to_json(cond.arguments[1], relation=cond.arguments[0])
    cond_json["operator"] = cond.name
    cond_json["arg0"] = cond_to_json(cond.arguments[0], relation=relation)
    if len(cond.arguments) == 2:
        cond_json['arg1'] = cond_to_json(cond.arguments[1], relation=relation)
    else:
        cond_json['arg1'] = [cond_to_json(a, relation=relation) for a in cond.arguments[1:]]
        json.dumps(cond_json)
    return cond_json


def key_to_json(key_arguments):
    assert len(key_arguments) == 1 or len(key_arguments) == 2
    if len(key_arguments) == 1:
        result = column_to_json(key_arguments)
    else:
        result = column_to_json(key_arguments[:1], relation=key_arguments[1])
    return result


def column_to_json(col, relation=None):
    cols = str(col[0]).strip('"').split(",")
    result = [dict(zip(['table', 'field'], c.split(".", maxsplit=1))) for c in cols]
    if relation is not None:
        for c in result:
            c['as'] = f"{c['table']}_{relation}"
    return result


def join_to_json(join_arguments):
    assert len(join_arguments) == 2 or len(join_arguments) == 3
    if len(join_arguments) == 2:
        return [column_to_json(c.arguments) for c in join_arguments]
    elif len(join_arguments) == 3:
        col0 = column_to_json(join_arguments[0].arguments)[0]
        col1 = column_to_json(join_arguments[1].arguments)[0]
        col1["as"] = f"{col1['table']}_{join_arguments[2]}"
        return [col0, col1]


def _parse_condition(condition):
    if condition.attrib['type'] in ["or", "and", "not"]:
        return "{}({})".format(condition.attrib['type'],
                               ",".join([_parse_condition(cnd)
                                         for cnd in condition.findall('condition')]))
    else:
        args = condition.findall('arg')
        arg_expr = [1] * len(args)
        for arg in args:
            if arg.attrib['type'] == "column":
                arg_expr[int(arg.attrib['pos'])] = 'column("{}")'.format(arg.attrib['val'])
            elif arg.attrib['type'] == "value":
                arg_expr[int(arg.attrib['pos'])] = 'value("{}")'.format(arg.attrib['val'])
        result = '{}({})'.format(condition.attrib['type'], ','.join(arg_expr))
        return (result)


def _parse_columnsmap(columnmap_xml):
    transf_function = columnmap_xml.attrib[
        'mappingFunction'] if "mappingFunction" in columnmap_xml.attrib else "identity"
    cols_names = set([column.attrib['name'] for column in columnmap_xml.findall("column")])
    columns = set(['column("{}")'.format(column) for column in cols_names])
    if len(columns) == 0:
        result = 'constant("{}")'.format(transf_function)
    else:
        result = 'apply("{}",{})'.format(transf_function, ','.join(columns))
    return cols_names, result


def _generate_condition_clauses(maps_to_onto, source_or_target):
    if source_or_target == "source":
        pred = "scond"
    else:
        pred = "tcond"
    lp = ''
    cond_atoms = dict()
    for onto_node in maps_to_onto:
        conds_for_onto_node = maps_to_onto[onto_node]['conditions']
        if len(conds_for_onto_node) > 0:
            cnd_atom = get_new_condition_atom("conjunction")
            for cnd in conds_for_onto_node:
                lp += f"{pred}({cnd}) :- {cnd_atom}.\n"
                lp += f"{pred}(has(R,{cnd})) :- has(R,({cnd_atom})).\n"
            pred_conds_for_onto_node = [f"{pred}({cnd})" for cnd in conds_for_onto_node]
            lp += f"{cnd_atom} :- {','.join(pred_conds_for_onto_node)}.\n"
            pred_has_R_conds_for_onto_node = [f"{pred}(has(R,{cnd}))" for cnd in conds_for_onto_node]
            lp += f"has(R,{cnd_atom}) :- {','.join(pred_has_R_conds_for_onto_node)}.\n"
            cond_atoms[onto_node] = cnd_atom
    return cond_atoms, lp


def get_new_condition_atom(cond_type):
    global _UNKN_CND_COUNTER, _CONJ_CND_COUNTER
    if cond_type == "conjunction":
        condition = "conj_cnd({})".format(_CONJ_CND_COUNTER)
        _CONJ_CND_COUNTER += 1
    if cond_type == "unknown":
        condition = "unknown_cnd({})".format(_UNKN_CND_COUNTER)
        _UNKN_CND_COUNTER += 1
    return condition


def compute_models(lp):
    ctrl = clingo.Control(["-Wno-atom-undefined"])
    ctrl.add("base", [], lp)
    ctrl.ground([("base", [])])
    ctrl.configuration.solve.models = "0"
    models = []
    for m in ctrl.solve(yield_=True):
        models.append(m.symbols(atoms=True))
    return models


def compute_mappings_from(models):
    model_cores_for_maps = defaultdict(list)
    models_for_maps = defaultdict(list)
    for m in models:
        model_core = get_core_from(m)
        maps_key = tuple(sorted(model_core['tmap'].union(model_core['cmap'])))
        model_cores_for_maps[maps_key].append(model_core)
        models_for_maps[maps_key].append(m)
    for maps_key in model_cores_for_maps:
        model_cores = model_cores_for_maps[maps_key]
        full_models = models_for_maps[maps_key]
        m_to_remove = []
        for i in range(len(model_cores)):
            if i not in m_to_remove:
                for j in range(i + 1, len(model_cores)):
                    if j not in m_to_remove:
                        if model_cores[i]['scond'].issubset(model_cores[j]['scond']) and \
                                model_cores[i]['cmap'].issuperset(model_cores[j]['cmap']):
                            # model_cores[i]['tcond'].issuperset(model_cores[j]['tcond']) and \
                            # model_cores[i]['tmap'].issuperset(model_cores[j]['tmap']) and \
                            m_to_remove.append(j)
                        elif model_cores[j]['scond'].issubset(model_cores[i]['scond']) and \
                                model_cores[j]['cmap'].issuperset(model_cores[i]['cmap']):
                            # model_cores[j]['tcond'].issuperset(model_cores[i]['tcond']) and \
                            # model_cores[j]['tmap'].issuperset(model_cores[i]['tmap']) and \
                            m_to_remove.append(i)
                            break
        model_cores_for_maps[maps_key] = [{
            "core": model_cores[i],
            "model": full_models[i]
            }
            for i in range(len(model_cores)) if i not in m_to_remove]
    return model_cores_for_maps


def get_mappings_from(m):
    return {c_or_t: {a for a in m if str(a)[:4] == c_or_t} for c_or_t in ['tmap', 'cmap']}


def get_core_from(m):
    return {op: {a for a in m if a.name == op} for op in ['scond', 'tcond', 'tmap', 'cmap']}


def get_conds_from(m):
    result = {s_or_t: {a for a in m if str(a)[:5] == s_or_t} for s_or_t in ['scond', 'tcond']}
    for s_or_t in ['scond', 'tcond']:
        result[f'{s_or_t[0]}table'] = [a.arguments[0].arguments[0] for a in result[s_or_t] if a.name == "in"]
    return result


if __name__ == "__main__":
    configs = configurations.Config("examples/SIB2SICRA/convert.ini")
    correspondences = Mappings(configs)
    correspondences.read_configs()
    tables_correspondences = [
        ['A01', 'A02', 'A04', 'A08', 'A10', 'A15', 'ANAGDI'],
        ['A01', 'A02', 'A04', 'A05', 'A08', 'Z81', 'ANAGDG'],
        ['A01', 'A02', 'A04', 'A05', 'A08', 'Z81', 'ANAGPF'],
        ['A01', 'A04', 'A08', 'A10', 'A15', 'ANAGPG']
        ]
    with mp.Pool(mp.cpu_count()) as pool:
        pool.map(correspondences.map_correspondences, tables_correspondences)
